from musicscore.musicxml.elements.note import Note, Grace, Duration, Beam, Tie, Type, Lyric
from musicscore.musicxml.elements.fullnote import Chord, Pitch, DisplayStep, DisplayOctave, Rest
from musicscore.dtd.dtd import ChildOccurrenceDTDConflict, ChildTypeDTDConflict, ChildIsNotOptional
from unittest import TestCase

from musicscore.musicxml.types.complextypes.lyric import Text


class TestNoteDTD(TestCase):
    def setUp(self):
        self.note = Note()

    def test_add_chord(self):

        ch = self.note.add_child(Chord())
        self.note.add_child(Pitch())
        self.note.add_child(Grace())
        with self.assertRaises(Exception):
            ch.add_child(Rest())

    def test_add_child_type(self):
        self.note.add_child(Pitch())
        self.note.add_child(Grace())
        with self.assertRaises(ChildTypeDTDConflict):
            self.note.add_child(Duration(1))

    def test_add_child_max_occurrence(self):
        self.note.add_child(Pitch())

        with self.assertRaises(ChildOccurrenceDTDConflict):
            self.note.add_child(Pitch())

    def test_close_1(self):
        self.note.add_child(Rest())
        self.note.add_child(Grace())
        self.note.close_dtd()
        result = ['Grace', 'Chord', 'Rest', 'Instrument', 'FootNote', 'Level', 'Voice', 'Type', 'Dot', 'Accidental',
                  'TimeModification', 'Stem', 'Notehead', 'NoteheadText', 'StaffElement', 'Beam', 'Notations', 'Lyric',
                  'Play']
        self.assertEqual([node.type_.__name__ for node in self.note.current_dtd_choice.traverse_leaves()], result)

    def test_close_2(self):
        self.note.add_child(Rest())
        with self.assertRaises(ChildIsNotOptional):
            self.note.close_dtd()

    def test_sort_children(self):
        self.note.add_child(Pitch())
        self.note.add_child(Beam('begin'))
        self.note.add_child(Tie())
        self.note.add_child(Beam('continue'))
        self.note.add_child(Duration(1))
        self.note.add_child(Tie())
        self.note.close_dtd()

        result = ['Pitch', 'Duration', 'Tie', 'Tie', 'Beam', 'Beam']
        self.assertEqual([type(child).__name__ for child in self.note.get_children()], result)

    def test_grace(self):
        self.note.add_child(Pitch())
        grace = Grace()
        grace.slash = 'yes'
        grace.make_time = 101
        with self.assertRaises(ValueError):
            grace.steal_time_following = 120

        grace.steal_time_following = 90
        self.note.add_child(grace)
        result = '''<note>
  <grace steal-time-following="90" make-time="101" slash="yes"/>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
</note>
'''
        self.assertEqual(self.note.to_string(), result)

#     def test_to_string(self):
#         self.note.add_child(Pitch())
#         self.note.add_child(Beam('begin'))
#         self.note.add_child(Tie())
#         self.note.add_child(Beam('continue'))
#         self.note.add_child(Duration(1))
#         self.note.add_child(Tie())
#         self.note.add_child(Chord())
#         self.note.close()
#         result = '''<note>
#   <chord/>
#   <pitch>
#     <step>C</step>
#     <octave>4</octave>
#   </pitch>
#   <duration>1</duration>
#   <tie/>
#   <tie/>
#   <beam number="1">begin</beam>
#   <beam number="1">continue</beam>
# </note>
# '''
#         self.assertEqual(self.note.to_string(), result)

    def test_rest(self):
        rest = self.note.add_child(Rest())
        self.note.add_child(Duration(1))
#         rest.add_child(DisplayOctave(4))
#         rest.add_child(DisplayStep('B'))
#         result = '''<note>
#   <rest>
#     <display-step>B</display-step>
#     <display-octave>4</display-octave>
#   </rest>
#   <duration>1</duration>
# </note>
# '''
#         self.assertEqual(self.note.to_string(), result)

    def test_type(self):
        self.note.add_child(Pitch())
        self.note.add_child(Duration())
        self.note.add_child(Type(value='quarter'))
        result = '''<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <type>quarter</type>
</note>
'''
        self.assertEqual(self.note.to_string(), result)

    def test_lyrics(self):
        self.note = Note()
        self.note.add_child(Pitch())
        self.note.add_child(Duration())
        self.note.add_child(Lyric()).add_child(Text('lyric 1'))
        self.note.add_child(Lyric(number='2')).add_child(Text('lyric 2'))
        result = '''<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <lyric number="1">
    <text>lyric 1</text>
  </lyric>
  <lyric number="2">
    <text>lyric 2</text>
  </lyric>
</note>
'''
        self.assertEqual(self.note.to_string(), result)



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from quicktions import Fraction
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(durations=[1.5, 1.5])
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_2(self):
        sf = SimpleFormat(durations=[2, 1.5])
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_2.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_3(self):
        sf = SimpleFormat(durations=[2.5, 1.5])
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_3.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_4(self):
        sf = SimpleFormat(durations=[1.5, 2.5])
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_4.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_5(self):
        sf = SimpleFormat(durations=[1.5, 0.5, 2], midis=[0, 0, 60])
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_5.xml'
        self.score.write(xml_path)
        # TestScore().assert_template(xml_path)

    def test_6(self):
        durations = [Fraction(1, 2), Fraction(5, 2), Fraction(1, 5), Fraction(4, 5)]
        sf = SimpleFormat(durations=durations, midis=[60, 0, 0, 60])
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_6.xml'
        self.score.write(xml_path)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class StartNote(AttributeAbstract):
    def __init__(self, start_note=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('start-note', start_note, "TypeStartNote")


class TrillStep(AttributeAbstract):
    def __init__(self, start_note=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('trill-step', start_note, "TypeTrillStep")


class TwoNoteTurn(AttributeAbstract):
    def __init__(self, two_note_turn=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('two-note-turn', two_note_turn, "TypeTwoNoteTurn")


class Accelerate(AttributeAbstract):
    def __init__(self, accelerate=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('accelerate', accelerate, "TypeYesNo")


class SecondBeat(AttributeAbstract):
    def __init__(self, second_beat=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('second-beat', second_beat, "TypePercent")


class LastBeat(AttributeAbstract):
    def __init__(self, last_beat=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('last-beat', last_beat, "TypePercent")


class Beats(AttributeAbstract):
    def __init__(self, beats=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('beats', beats, "TypeTrillBeats")


class TrillSound(StartNote, TrillStep, TwoNoteTurn, Accelerate, Beats, SecondBeat, LastBeat):
    """
    The trill-sound attribute group includes attributes used to guide the sound of trills, mordents, turns, shakes, and
    wavy lines, based on MuseData sound suggestions. The default choices are:

                start-note = "upper"
                trill-step = "whole"
                two-note-turn = "none"
                accelerate = "no"
                beats = "4".

                Second-beat and last-beat are percentages for landing on the indicated beat, with defaults of 25 and 75
                respectively.

                For mordent and inverted-mordent elements, the defaults are different:

                The default start-note is "main", not "upper".
                The default for beats is "3", not "4".
                The default for second-beat is "12", not "25".
                The default for last-beat is "24", not "75".
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class TimeOnly(AttributeAbstract):
    def __init__(self, time_only=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('time-only', time_only, 'TypeTimeOnly')


class XMLError(Exception):
    pass


class AfterInitializationError(XMLError):
    def __init__(self, tag):
        msg = 'tag {} cannot be changed after initializing XMLElement'.format(tag)
        super().__init__(msg)


class XMLTypeError(XMLError):
    def __init__(self, allowed_types):
        msg = 'only {} are allowed'.format(allowed_types)
        super().__init__(msg)


class XMLValueError(XMLError):
    def __init__(self, allowed_values):
        msg = 'only {} are allowed'.format(allowed_values)
        super().__init__(msg)


class ChildAlreadyExists(XMLError):
    def __init__(self, new_child):
        msg = 'not multiple type of child {} already exists'.format(type(new_child))
        super().__init__(msg)



import re
import functools
from itertools import cycle, islice


def replace_dash(name):
    if '-' in name:
        return re.sub(r'-', '_', name)
    elif '_' in name:
        return re.sub(r'_', '-', name)
    else:
        return name


def is_empty(string):
    if re.sub(r'\s', '', string) == '':
        return True
    else:
        return False


def lcm(l):
    """least common multiple of numbers in a list"""

    def _lcm(a, b):
        if a > b:
            greater = a
        else:
            greater = b

        while True:
            if greater % a == 0 and greater % b == 0:
                lcm_ = greater
                break
            greater += 1

        return lcm_

    return functools.reduce(lambda x, y: _lcm(x, y), l)


def flatten(x):
    if hasattr(x, "__iter__"):
        result = []
        for el in x:
            if hasattr(el, "__iter__") and not isinstance(el, str):
                result.extend(flatten(el))
            else:
                result.append(el)
        return result
    else:
        return [x]


def roundrobin(*iterables):
    "roundrobin('ABC', 'D', 'EF') --> A D E B F C"
    # Recipe credited to George Sakkis
    num_active = len(iterables)
    nexts = cycle(iter(it).__next__ for it in iterables)
    while num_active:
        try:
            for next in nexts:
                yield next()
        except StopIteration:
            num_active -= 1
            nexts = cycle(islice(nexts, num_active))


def substitute(input_list, old_element, new_elements):
    index = input_list.index(old_element)
    try:
        new_elements = list(new_elements)
    except TypeError:
        new_elements = [new_elements]

    output_list = input_list[:index] + new_elements + input_list[index + 1:]
    return output_list


def dToX(input_list: object, first_element: object = 0) -> object:
    if isinstance(input_list, list) is False:
        raise TypeError('xToD(input_list)')
    else:
        output = [first_element]
        for i in range(len(input_list)):
            output.append(input_list[i] + output[i])
        return output


def xToD(input_list):
    result = []
    for i in range(1, len(input_list)):
        result.append(input_list[i] - input_list[i - 1])
    return result


def step_sums(input):
    # if input == {}:
    #     return {}
    if not input:
        return []

    # if isinstance(input, dict):
    #     input_items = zip(*input.items())[1]
    #     input_keys = zip(*input.items())[0]
    #     output_items = [input_items[0]]
    #     for i in range(1, len(input_items)):
    #         output_items.append(output_items[-1] + input_items[i])
    #
    #     return dict(zip(input_keys, output_items))

    result = [input[0]]
    for i in range(1, len(input)):
        result.append(result[-1] + input[i])
    return result


def one_dimensional(x):
    if hasattr(x, '__iter__'):
        result = []
        for list_el in x:
            if hasattr(list_el, "__iter__") and not hasattr(list_el[0], "__iter__"):
                result.append(list_el)
            else:
                result.extend(one_dimensional(list_el))
        return result
    else:
        return [x]



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()
        sf = SimpleFormat(durations=30 * [4])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

    def test_1(self):
        self.score.page_style.orientation = 'landscape'

        for index, measure in enumerate(self.score.get_children_by_type(TreeMeasure)):
            if index % 4 == 0:
                measure.add_system_break()

        result_path = path + '_test_1'
        self.score.finish()
        partwise = self.score.to_partwise()
        partwise.write(path=result_path)
        # self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from unittest import TestCase

from musicscore.musicxml.groups.musicdata import Direction
from musicscore.musicxml.types.complextypes.direction import DirectionType
from musicscore.musicxml.types.complextypes.directiontype import Bracket


class Test(TestCase):

    def test_1(self):
        direction = Direction()
        direction_type = direction.add_child(DirectionType())
        direction_type.add_child(Bracket(type='start', line_end='arrow'))

        result = """<direction>
  <direction-type>
    <bracket number="1" type="start" line-end="arrow"/>
  </direction-type>
</direction>
"""
        self.assertEqual(direction.to_string(), result)



from unittest import TestCase

from musicscore.musictree.exceptions import MusicTreeError
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musictree.treetime import TreeTime


class TestTreeTime(TestCase):

    def test_tree_time(self):
        time = TreeTime(3, 4, 1, 8)
        result = '''<time>
  <beats>3</beats>
  <beat-type>4</beat-type>
  <beats>1</beats>
  <beat-type>8</beat-type>
</time>
'''
        self.assertEqual(time.to_string(), result)

        # self.assertEqual(time.quarter_duration, 3.5)

        with self.assertRaises(MusicTreeError):
            TreeTime(4, 3)

    def test_copy_time(self):
        time = TreeTime(3, 4, 1, 8)
        copy = time.__copy__()
        result = '''<time>
  <beats>3</beats>
  <beat-type>4</beat-type>
  <beats>1</beats>
  <beat-type>8</beat-type>
</time>
'''
        self.assertEqual(copy.to_string(), result)



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.note import Tie


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        simpleformat = SimpleFormat(durations=4)
        voice = simpleformat.to_stream_voice(2)
        self.score.add_part()
        self.score.add_measure()
        p = self.score.get_measure(1).get_part(1)
        remaining_chords = voice.add_to_part(p)
        result = '''<part id="p1">
  <attributes>
    <divisions>1</divisions>
  </attributes>
  <note>
    <pitch>
      <step>B</step>
      <octave>4</octave>
    </pitch>
    <duration>4</duration>
    <voice>2</voice>
    <type>whole</type>
  </note>
</part>
'''
        self.assertEqual(p.to_string(), result)
        self.assertEqual(remaining_chords, None)

    def test_2(self):
        simpleformat = SimpleFormat(durations=7)
        voice = simpleformat.to_stream_voice(2)
        self.score.add_part()
        self.score.add_measure()
        p = self.score.get_measure(1).get_part(1)
        remaining_chords = voice.add_to_part(p)
        result = '''<part id="p1">
  <attributes>
    <divisions>1</divisions>
  </attributes>
  <note>
    <pitch>
      <step>B</step>
      <octave>4</octave>
    </pitch>
    <duration>4</duration>
    <tie type="start"/>
    <voice>2</voice>
    <type>whole</type>
    <notations>
      <tied number="1" type="start"/>
    </notations>
  </note>
</part>
'''
        self.assertEqual(p.to_string(), result)
        chord = remaining_chords[0]
        self.assertEqual(chord.quarter_duration, 3)
        self.assertEqual(chord.get_children_by_type(Tie)[0].type, 'stop')

    def test_3(self):
        simpleformat = SimpleFormat(durations=3)
        voice = simpleformat.to_stream_voice(2)
        self.score.add_part()
        self.score.add_measure()
        p = self.score.get_measure(1).get_part(1)
        remaining_chords = voice.add_to_part(p)
        result = '''<part id="p1">
  <attributes>
    <divisions>1</divisions>
  </attributes>
  <note>
    <pitch>
      <step>B</step>
      <octave>4</octave>
    </pitch>
    <duration>3</duration>
    <voice>2</voice>
    <type>half</type>
    <dot/>
  </note>
  <note>
    <rest/>
    <duration>1</duration>
    <voice>2</voice>
    <type>quarter</type>
  </note>
</part>
'''
        self.assertEqual(p.to_string(), result)

    def test_4(self):
        simpleformat = SimpleFormat(durations=7, midis=0)
        voice = simpleformat.to_stream_voice(2)
        self.score.add_part()
        self.score.add_measure()
        p = self.score.get_measure(1).get_part(1)
        remaining_chords = voice.add_to_part(p)
        result = '''<part id="p1">
  <attributes>
    <divisions>1</divisions>
  </attributes>
  <note>
    <rest/>
    <duration>4</duration>
    <voice>2</voice>
    <type>whole</type>
  </note>
</part>
'''
        self.assertEqual(p.to_string(), result)
        self.assertEqual(remaining_chords[0].midis[0].value, 0)
        self.assertEqual(remaining_chords[0].get_children_by_type(Tie), [])

    def test_5(self):
        simpleformat = SimpleFormat(durations=3, midis=0)
        voice = simpleformat.to_stream_voice(2)
        self.score.add_part()
        self.score.add_measure()
        p = self.score.get_measure(1).get_part(1)
        voice.add_to_part(p)
        result = '''<part id="p1">
  <attributes>
    <divisions>1</divisions>
  </attributes>
  <note>
    <rest/>
    <duration>4</duration>
    <voice>2</voice>
    <type>whole</type>
  </note>
</part>
'''
        self.assertEqual(p.to_string(), result)



from musicscore.musicxml.attributes.attribute_abstract import String
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.font import Font
from musicscore.musicxml.attributes.letterspacing import LetterSpacing
from musicscore.musicxml.attributes.text_direction import TextDirection
from musicscore.musicxml.attributes.textdecoration import TextDecoration
from musicscore.musicxml.attributes.textrotation import TextRotation
from musicscore.musicxml.types.complextypes.complextype import ComplexType


# todo XMLLang
class ComplexTypeTextElementData(ComplexType, String, Font, Color, TextDecoration, TextRotation, LetterSpacing,
                                 TextDirection):
    """
    documentation>The text-element-data type represents a syllable or portion of a syllable for lyric text underlay. A
    hyphen in the string content should only be used for an actual hyphenated word. Language names for text elements
    come from ISO 639, with optional country subcodes from ISO 3166.
    """

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)



from musicscore.musictree.treeinstruments import TreeInstrument
from musicscore.musictree.treepart import TreePart
from musicscore.musicxml.types.complextypes.namedisplay import DisplayText
from musicscore.musicxml.types.complextypes.partgroup import GroupNameDisplay, GroupSymbol, GroupBarline
from musicscore.musicxml.types.complextypes.partlist import ScorePart, PartGroup

from musicscore.musicxml.types.complextypes.scorepart import PartName, PartAbbreviation


class TreeScorePart(ScorePart):

    def __init__(self, id, instrument=None, *args, **kwargs):
        super().__init__(id=id, *args, **kwargs)
        self.add_child(PartName(name='none', print_object='no'))
        self._instrument = None
        self.instrument = instrument
        self._max_division = None
        self._forbidden_divisions = None
        self._parts = []
        self.parent_score = None

    @property
    def part_name(self):
        try:
            return self.get_children_by_type(PartName)[0]
        except IndexError:
            return None

    @property
    def part_abbreviation(self):
        try:
            return self.get_children_by_type(PartAbbreviation)[0]
        except IndexError:
            return None

    @property
    def instrument(self):
        return self._instrument

    @instrument.setter
    def instrument(self, val):
        if val is not None and not isinstance(val, TreeInstrument):
            raise TypeError('instrument.value must be of type TreeInstrument not{}'.format(type(val)))
        self._instrument = val
        if val is not None:
            val.id = self.id
            if self.part_name is not None:
                self.remove_child(self.part_name)
            self.add_child(self.instrument.part_name)

            if self.part_abbreviation is not None:
                self.remove_child(self.part_abbreviation)
            self.add_child(self.instrument.part_abbreviation)

    @property
    def max_division(self):
        if self._max_division is None:
            self._max_division = self.parent_score.max_division
        return self._max_division

    @max_division.setter
    def max_division(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError('max_division.value must be None or of type int not {}'.format(type(value)))

        self._max_division = value

    @property
    def forbidden_divisions(self):
        if self._forbidden_divisions is None:
            self._forbidden_divisions = self.parent_score.forbidden_divisions

        return self._forbidden_divisions

    @forbidden_divisions.setter
    def forbidden_divisions(self, value):
        if value is not None:
            for x in value:
                if not isinstance(x, int):
                    raise TypeError('forbidden_division must be of type int not{}'.format(type(value)))

        self._forbidden_divisions = value

    def get_parts(self):
        return self._parts

    def add_part(self, part=None):
        if not part:
            part = TreePart(id=self.id)
        else:
            if part.id != self.id:
                raise ValueError('Part must have the same id as TreeScorePart')
        part.parent_score_part = self
        self._parts.append(part)
        return part

    def add_part_group(self, number, type, name=None, symbol=None, barline=None):
        pg = PartGroup(type=type, number=str(number))
        pg._up = self.up
        if name:
            gnd = pg.add_child(GroupNameDisplay())
            gnd.add_child(DisplayText(name))
        if symbol:
            pg.add_child(GroupSymbol(symbol))
        if barline:
            pg.add_child(GroupBarline(barline))
        self_index = self.up.current_children.index(self)
        if type == 'start':
            self.up.current_children.insert(self_index, pg)
        else:
            self.up.current_children.insert(self_index + 1, pg)



from musicscore.dtd.dtd import Sequence, Element, Choice
from musicscore.musicxml.attributes.id import Id
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType, Empty
from musicscore.musicxml.types.complextypes.virtualinstrument import ComplexTypeVirtualInstrument
from musicscore.musicxml.types.simple_type import String, TypePositiveIntegerOrEmpty


class InstrumentName(XMLElement, String):
    """
    The instrument-name element is typically used within a software application, rather than appearing on the printed
    page of a score.
    """
    _TAG = 'instrument-name'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class InstrumentAbbreviation(XMLElement, String):
    """
    The optional instrument-abbreviation element is typically used within a software application, rather than appearing
    on the printed page of a score.
    """
    _TAG = 'instrument-abbreviation'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class InstrumentSound(XMLElement, String):
    """
    The instrument-sound element describes the default timbre of the score-instrument.
    This description is independent of a particular virtual or MIDI instrument specification and allows playback to be
    shared more easily between applications and libraries.
    """
    _TAG = 'instrument-sound'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Solo(Empty):
    """
    The solo element was added in Version 2.0. It is present if performance is intended by a solo instrument.
    """
    _TAG = 'solo'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Ensemble(XMLElement, TypePositiveIntegerOrEmpty):
    """
    The ensemble element was added in Version 2.0. It is present if performance is intended by an ensemble such as an
    orchestral section. The text of the ensemble element contains the size of the section, or is empty if the ensemble
    size is not specified
    """
    _TAG = 'ensemble'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class VirtualInstrument(ComplexTypeVirtualInstrument):
    _TAG = 'virtual-instrument'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeScoreInstrument(ComplexType, Id):
    """
    The score-instrument type represents a single instrument within a score-part. As with the score-part type,
    each score-instrument has a required ID attribute, a name, and an optional  abbreviation.

    A score-instrument type is also required if the score specifies MIDI 1.0 channels, banks, or programs.
    An initial midi-instrument assignment can also be made here. MusicXML software should be able to automatically
    assign reasonable channels and instruments without these elements in simple cases, such as where part names match
    General MIDI instrument names.
    """
    _DTD = Sequence(
        Element(InstrumentName, min_occurrence=0),
        Element(InstrumentAbbreviation, min_occurrence=0),
        Element(InstrumentSound, min_occurrence=0),
        Choice(
            Element(Solo),
            Element(Ensemble),
            min_occurrence=0
        ),
        Element(VirtualInstrument, min_occurrence=0)
    )

    def __init__(self, tag, id, *args, **kwargs):
        super().__init__(tag=tag, id=id, *args, **kwargs)



from musicscore.dtd.dtd import Sequence, Element, Choice
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.types.complextypes.accidentalmark import ComplexTypeAccidentalMark
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.emptytrillsound import ComplexTypeEmptyTrillSound
from musicscore.musicxml.types.complextypes.horizontalturn import ComplexTypeHorizontalTurn
from musicscore.musicxml.types.complextypes.tremolo import ComplexTypeTremolo
from musicscore.musicxml.types.complextypes.wavyline import ComplexTypeWavyLine


class TrillMark(ComplexTypeEmptyTrillSound):
    """
    The trill-mark element represents the trill-mark symbol.
    """
    _TAG = "trill-mark"

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Turn(ComplexTypeHorizontalTurn):
    """
    The turn element is the normal turn shape which goes up then down.
    """
    _TAG = "turn"

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class DelayedTurn(ComplexTypeHorizontalTurn):
    """
    The delayed-turn element indicates a normal turn that is delayed until the end of the current note.
    """
    _TAG = "delayed-turn"

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class InvertedTurn(ComplexTypeHorizontalTurn):
    """
    The inverted-turn element has the shape which goes down and then up.
    """
    _TAG = "inverted-turn"

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class DelayedInvertedTurn(ComplexTypeHorizontalTurn):
    """
    The delayed-inverted-turn element indicates an inverted turn that is delayed until the end of the current note.
    """
    _TAG = "delayed-inverted-turn"

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class VerticalTurn(ComplexTypeEmptyTrillSound):
    """
    The vertical-turn element has the turn symbol shape arranged vertically going from upper left to lower right.
    """
    _TAG = "vertical-turn"

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class InvertedVerticalTurn(ComplexTypeEmptyTrillSound):
    """
    The inverted-vertical-turn element has the turn symbol shape arranged vertically going from upper right to lower
    left.
    """
    _TAG = "inverted-vertical-turn"

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Shake(ComplexTypeEmptyTrillSound):
    """
    The shake element has a similar appearance to an inverted-mordent element.
    """
    _TAG = "shake"

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class WavyLine(ComplexTypeWavyLine):
    _TAG = 'wavy-line'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Mordent(object):
    """
                    <xs:element name="mordent" type="mordent">
                    <xs:annotation>
                        <xs:documentation>The mordent element represents the sign with the vertical line. The choice of
                            which mordent sign is inverted differs between MusicXML and SMuFL. The long attribute is
                            "no" by default.
                        </xs:documentation>
                    </xs:annotation>
                </xs:element>
    """
    _TAG = "mordent"

    def __init__(self, *args, **kwargs):
        super().__init__(self, tag=self._TAG, *args, **kwargs)
        raise NotImplementedError()


class InvertedMordent(object):
    """
                    <xs:element name="inverted-mordent" type="mordent">
                    <xs:annotation>
                        <xs:documentation>The inverted-mordent element represents the sign without the vertical line.
                            The choice of which mordent is inverted differs between MusicXML and SMuFL. The long
                            attribute is "no" by default.
                        </xs:documentation>
                    </xs:annotation>
                </xs:element>
    """
    _TAG = "inverted-mordent"

    def __init__(self, *args, **kwargs):
        super().__init__(self, tag=self._TAG, *args, **kwargs)
        raise NotImplementedError()


class Schleifer(object):
    """
                    <xs:element name="schleifer" type="empty-placement">
                    <xs:annotation>
                        <xs:documentation>The name for this ornament is based on the German, to avoid confusion with the
                            more common slide element defined earlier.
                        </xs:documentation>
                    </xs:annotation>
                </xs:element>
    """
    _TAG = "schleifer"

    def __init__(self, *args, **kwargs):
        super().__init__(self, tag=self._TAG, *args, **kwargs)
        raise NotImplementedError()


class Tremolo(ComplexTypeTremolo):
    _TAG = 'tremolo'

    def __init__(self, value=3, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Haydn(ComplexTypeEmptyTrillSound):
    """
    The haydn element represents the Haydn ornament. This is defined in SMuFL as ornamentHaydn.
    """
    _TAG = "haydn"

    def __init__(self, *args, **kwargs):
        super().__init__(self, tag=self._TAG, *args, **kwargs)


class OtherOrnament(object):
    """
                    <xs:element name="other-ornament" type="other-placement-text">
                    <xs:annotation>
                        <xs:documentation>The other-ornament element is used to define any ornaments not yet in the
                            MusicXML format. The smufl attribute can be used to specify a particular ornament, allowing
                            application interoperability without requiring every SMuFL ornament to have a MusicXML
                            element equivalent. Using the other-ornament element without the smufl attribute allows for
                            extended representation, though without application interoperability.
                        </xs:documentation>
                    </xs:annotation>
                </xs:element>
    """
    _TAG = "other-ornament"

    def __init__(self, *args, **kwargs):
        super().__init__(self, tag=self._TAG, *args, **kwargs)
        raise NotImplementedError()


class AccidentalMark(ComplexTypeAccidentalMark):
    _TAG = "accidental-mark"

    def __init__(self, *args, **kwargs):
        super().__init__(self, tag=self._TAG, *args, **kwargs)
        raise NotImplementedError()


class ComplexTypeOrnaments(ComplexType, OptionalUniqueId):
    """
    Ornaments can be any of several types, followed optionally by accidentals. The accidental-mark element's content is
    represented the same as an accidental element, but with a different name to reflect the different musical meaning.
    """

    _DTD = Sequence(
        Choice(
            Element(TrillMark),
            Element(Turn),
            Element(DelayedTurn),
            Element(InvertedTurn),
            Element(DelayedInvertedTurn),
            Element(VerticalTurn),
            Element(InvertedVerticalTurn),
            Element(Shake),
            Element(WavyLine),
            Element(Mordent),
            Element(InvertedMordent),
            Element(Schleifer),
            Element(Tremolo),
            Element(Haydn),
            Element(OtherOrnament)
            , min_occurrence=0, max_occurrence=None
        ),
        Element(AccidentalMark,
                min_occurrence=0, max_occurrence=None
                )

    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



import warnings

from lxml import etree as et
from quicktions import Fraction

from musicscore.basic_functions import lcm, substitute, flatten
from musicscore.musictree.exceptions import MusicTreeError
from musicscore.musictree.treebeat import TreeBeat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treechordflags2 import TreeChordFlag2
from musicscore.musictree.treechordflags3 import TreeChordFlag3
from musicscore.musictree.treenote import TreeNote, TreeBackup
from musicscore.musicxml.elements import timewise as timewise
from musicscore.musicxml.elements.fullnote import Pitch, Rest
from musicscore.musicxml.elements.note import Beam, Type, Tie, Notations
from musicscore.musicxml.groups.common import Voice
from musicscore.musicxml.groups.musicdata import Direction, Attributes
from musicscore.musicxml.types.complextypes.attributes import Divisions
from musicscore.musicxml.types.complextypes.direction import DirectionType, Sound
from musicscore.musicxml.types.complextypes.directiontype import Metronome
from musicscore.musicxml.types.complextypes.metronome import BeatUnit, PerMinute
from musicscore.musicxml.types.complextypes.notations import Slur


class XMLChord(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._notes = []
        self._parent_voice = None

    def add_note(self, note):
        if self.notes:
            if note.offset != self.offset:
                raise ValueError('notes of XMLChord must have the same offset')
            note_voice_number = note.get_children_by_type(Voice)[0].value

            if note_voice_number != self.voice_number:
                raise ValueError('notes of XMLChord must have the same voice number {} and {}'.format(note_voice_number,
                                                                                                      self.voice_number))
        self._notes.append(note)

    @property
    def notes(self):
        return self._notes

    @property
    def parent_voice(self):
        return self._parent_voice

    @parent_voice.setter
    def parent_voice(self, value):
        if not isinstance(value, TreePartVoice):
            raise TypeError()
        self._parent_voice = value

    @property
    def offset(self):
        if self.notes:
            return self.notes[-1].offset
        else:
            return None

    @property
    def voice_number(self):
        if self.notes:
            return self.notes[-1].get_children_by_type(Voice)[0].value

    @property
    def previous(self):
        index = self.parent_voice.xml_chords.index(self)
        if index == 0:
            return None
        return self.parent_voice.xml_chords[index - 1]

    @property
    def is_tied_to_previous(self):
        for note in self.notes:
            if 'stop' in [t.type for t in note.get_children_by_type(Tie)]:
                return True
        return False

    def has_same_pitches(self, xml_chord):
        if len(self.notes) == len(xml_chord.notes):
            for note_1, note_2 in zip(self.notes, xml_chord.notes):
                if isinstance(note_1.event, Rest) or isinstance(note_2.event, Rest):
                    return False
                pitch_1 = note_1.pitch.dump()
                pitch_2 = note_2.pitch.dump()
                if len(pitch_1) == len(pitch_2):
                    for i in range(1, len(pitch_1)):
                        if pitch_1[i].value != pitch_2[i].value:
                            return False
                    return True
        return False


class TreePartVoice(object):
    def __init__(self, number=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._number = None
        self.number = number
        self._max_division = None
        self._forbidden_divisions = None
        self._chords = []
        self._beats = None
        self._filled_with_rest = False
        self._preliminary_rests_adjoined = False
        self._beats_added = False
        self._quantized = False
        self._flags_implemented = False
        self._not_notatable_split = False
        self._ties_adjoined = False
        self._rests_adjoined = False
        self._tuplets_updated = False
        self._sextoles_substituted = False
        self._types_updated = False
        self._dots_updated = False
        self._flags2_implemented = False
        self._flags3_implemented = False
        self.parent_part = None
        self._xml_chords = None

    @property
    def __name__(self):
        # return self.parent_part.__name__ + ' ' + 'v:' + str(self.number)

        return self.parent_part.__name__ + '.' + str(self.number)

    @property
    def chords(self):
        return self._chords

    @property
    def number(self):
        return self._number

    @number.setter
    def number(self, value):
        if not isinstance(value, int):
            raise TypeError('number.value must be of type int not{}'.format(type(value)))
        self._number = value

    @property
    def max_division(self):
        if self._max_division is None:
            self._max_division = self.parent_part.max_division
        return self._max_division

    @max_division.setter
    def max_division(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError('max_division.value must be None or of type int not {}'.format(type(value)))

        self._max_division = value

    @property
    def forbidden_divisions(self):
        if self._forbidden_divisions is None:
            self._forbidden_divisions = self.parent_part.forbidden_divisions

        return self._forbidden_divisions

    @forbidden_divisions.setter
    def forbidden_divisions(self, value):
        if value is not None:
            for x in value:
                if not isinstance(x, int):
                    raise TypeError('forbidden_division must be of type int not{}'.format(type(value)))

        self._forbidden_divisions = value

    @property
    def next(self):
        parts = self.parent_part.parent_score_part.get_parts()
        index = parts.index(self.parent_part)
        if index < len(parts):
            next_part = parts[index + 1]
        else:
            return None
        try:
            return next_part.get_voice(self.number)
        except KeyError:
            return None

    @property
    def previous(self):
        parts = self.parent_part.parent_score_part.get_parts()
        index = parts.index(self.parent_part)
        if index > 0:
            previous_part = parts[index - 1]
        else:
            return None
        try:
            return previous_part.get_voice(self.number)
        except KeyError:
            return None

    def add_chord(self, chord):
        # print('treepart add_chord, chord.tie_orientation', chord.tie_orientation)
        remain = chord.quarter_duration - self.remaining_duration
        if self.remaining_duration == 0:
            return chord

        elif remain > 0:
            split = chord.split([chord.quarter_duration - remain, remain])
            first_chord = split[0]
            self.chords.append(first_chord)
            first_chord.parent_voice = self
            if first_chord.manual_voice_number:
                first_chord.add_child(Voice(str(first_chord.manual_voice_number)))
            else:
                first_chord.add_child(Voice(str(self.number)))
            first_chord._head = True
            split[1]._tail = True
            return split[1]
        else:
            self.chords.append(chord)
            chord.parent_voice = self
            if chord.manual_voice_number:
                chord.add_child(Voice(str(chord.manual_voice_number)))
            else:
                chord.add_child(Voice(str(self.number)))

    def remove_chords(self):
        self._chords = []

    @property
    def remaining_duration(self):
        measure = self.parent_part.up
        if self.chords:
            return measure.quarter_duration - self.chords[-1].end_position
        return measure.quarter_duration

    @property
    def xml_chords(self):
        if not self._xml_chords:
            xml_chords = []
            for note in [note for note in self.parent_part.get_children_by_type(TreeNote)]:
                offset_voices = [(xml_chord.offset, xml_chord.voice_number) for xml_chord in xml_chords]
                note_offset_voice = (note.offset, note.get_children_by_type(Voice)[0].value)
                if note_offset_voice in offset_voices:
                    index = offset_voices.index(note_offset_voice)
                    xml_chord = xml_chords[index]
                    xml_chord.add_note(note)
                else:
                    xml_chord = XMLChord()
                    xml_chord.parent_voice = self
                    xml_chord.add_note(note)
                    xml_chords.append(xml_chord)
            self._xml_chords = xml_chords

        return self._xml_chords

    def _group_beats(self, grouping_list):
        # todo test grouping list
        if False:
            raise Exception()
        else:
            chords = [chord for chord in self.chords if not chord.is_finger_tremolo]
            grouped_chords = []
            group_positions = [0]
            for group in grouping_list:
                # todo list of beat_types

                beat_type = self.parent_part.up.time.beat_type
                if isinstance(beat_type, list):
                    raise NotImplementedError()

                group_positions.append(group_positions[-1] + group * (4. / beat_type.value))
            current_chord_index = 0
            group_positions.pop(0)
            for group_position in group_positions:
                current_grouped_chords = []
                while current_chord_index < len(chords) and chords[current_chord_index].offset < group_position:
                    current_grouped_chords.append(chords[current_chord_index])
                    current_chord_index = current_chord_index + 1

                grouped_chords.append(current_grouped_chords)
            return grouped_chords

    def group_beams(self, grouping_list=None):

        def _generate_grouping_list():
            grouping_list = []
            for (beats, beat_type) in self.parent_part.up.time.get_time_signatures():
                if beats.value % 3 == 0 and beat_type.value != 4:
                    for x in range(beats.value // 3):
                        grouping_list.append(3)
                else:
                    for x in range(beats.value):
                        grouping_list.append(1)
            return grouping_list

        def _set_beams(grouped_chords):

            global begin_beam, continue_end_beam

            def add_beam(chord, chord_group, chord_group_index, number):
                global begin_beam, continue_end_beam
                beam = chord.add_child(Beam(None, number=number))
                if chord_group_index != len(chord_group) - 1:
                    if begin_beam:
                        beam.value = 'begin'
                    elif continue_end_beam:
                        if chord_group_index == len(chord_group) - 1:
                            beam.value = 'end'
                        else:
                            beam.value = 'continue'
                elif len(chord_group) > 1 and chord_group[chord_group_index - 1] != []:
                    beam.value = 'end'

            for chord_group in grouped_chords:
                begin_beam = True
                continue_end_beam = False

                for i in range(len(chord_group)):
                    chord = chord_group[i]
                    type = chord.get_children_by_type(Type)[0].value
                    numbers = {'eighth': 1, '16th': 2, '32nd': 3, '64th': 4, '128th': 5}
                    if type in numbers and chord.quarter_duration != 0:
                        for index in range(numbers[type]):
                            add_beam(chord, chord_group, i, number=index + 1)
                        if begin_beam:
                            begin_beam = False
                            continue_end_beam = True
                        elif continue_end_beam and i == len(chord_group) - 1:
                            begin_beam = True
                            continue_end_beam = False
                        # beam = chord.add_child(Beam(None))
                        # if i != len(group) - 1:
                        #     if begin_beam:
                        #         beam.value = 'begin'
                        #         begin_beam = False
                        #         continue_end_beam = True
                        #
                        #     elif continue_end_beam:
                        #         if i == len(group) - 1:
                        #             beam.value = 'end'
                        #             begin_beam = True
                        #             continue_end_beam = False
                        #         else:
                        #             beam.value = 'continue'
                        # elif len(group) > 1 and group[i - 1] != []:
                        #     beam.value = 'end'

                    elif i != 0:
                        # print('treepart group_beams, elif')
                        try:
                            beam = chord_group[i - 1].beam

                            if beam.value == 'begin':
                                chord_group[i - 1].remove_child(beam)
                                begin_beam = True
                                continue_end_beam = False
                            elif beam.value == 'continue':
                                chord_group[i - 1].beam.value = 'end'
                                begin_beam = True
                                continue_end_beam = False
                        except AttributeError:
                            pass

        if grouping_list is None:
            grouping_list = _generate_grouping_list()
            grouped_chords = self._group_beats(grouping_list)
            _set_beams(grouped_chords)
        else:
            raise NotImplementedError('group_beams with values other than None')

    def set_beats(self, list_of_beats=None):
        if not self.parent_part.up:
            raise Exception('voice must have a part as child of a measure to be able to set beats')

        if list_of_beats is None:
            list_of_beats = []
            for time_signature in self.parent_part.up.time.get_time_signatures():
                (beats, beat_type) = time_signature
                for b in range(beats.value):
                    tree_beat = TreeBeat(duration=4. / beat_type.value)
                    list_of_beats.append(tree_beat)
                    tree_beat.parent_voice = self
        else:
            duration = 0
            for beat in list_of_beats:
                beat.parent_voice = self
                duration += beat.duration
            if self.parent_part.up.quarter_duration != duration:
                raise ValueError('sum of beat durations must be equal to measure duration')

        self._beats = list_of_beats

    @property
    def beats(self):
        return self._beats

    def get_beats(self):
        return self.beats

    def _add_chords_to_beats(self):
        if len(self.beats) == 1:
            current_beat = self.beats[0]
            for chord in self.chords:
                current_beat.add_chord(chord)
        else:
            beats = iter(self.beats)
            current_beat = beats.__next__()
            next_beat = beats.__next__()

            for chord in self.chords:
                while True and (chord.offset < current_beat.offset or chord.offset >= next_beat.offset):
                    try:
                        current_beat = next_beat
                        next_beat = beats.__next__()
                    except StopIteration:
                        break

                current_beat.add_chord(chord)

    def _split_chords_beatwise(self):

        for beat in self.beats:
            if beat.chords:
                first_chord = beat.chords[0]
                if beat.offset < first_chord.offset:
                    previous_chord = first_chord.previous
                    tail_duration = (previous_chord.end_position - beat.offset)
                    ratios = [previous_chord.quarter_duration - tail_duration, tail_duration]

                    split = previous_chord.split(ratios)
                    self._chords = substitute(self._chords, previous_chord, split)
                    beat.chords.insert(0, split[1])
                    split[1].parent_beat = beat
                    split[0]._head = True
                    split[1]._tail = True
                    # print([chord.quarter_duration for chord in split])

        for beat in self.beats:
            if beat.chords and len([chord for chord in beat.chords if chord.quarter_duration != 0]) != 1:
                last_chord = beat.chords[-1]
                if last_chord.end_position > beat.end_position:
                    head_duration = (beat.end_position - last_chord.offset)
                    ratios = [head_duration, last_chord.quarter_duration - head_duration]
                    # print(ratios)
                    split = last_chord.split(ratios)
                    self._chords = substitute(self._chords, last_chord, split)
                    beat.next.chords.insert(0, split[1])
                    split[1].parent_beat = beat.next
                    split[0]._head = True
                    split[1]._tail = True

    def _correct_deviations(self):

        def _get_expected_durations():
            output = [self.beats[0].duration]
            for beat in self.beats[1:]:
                if beat.chords == []:
                    output[-1] += beat.duration
                else:
                    output.append(beat.duration)
            return output

        expected_durations = _get_expected_durations()
        for index, beat in enumerate([beat for beat in self.beats if beat.chords]):
            delta = expected_durations[index] - sum([chord.quarter_duration for chord in beat.chords])
            beat.chords[-1].quarter_duration += delta

    def fill_with_rest(self):
        if not self._filled_with_rest:
            if self.remaining_duration > 0:
                if self.chords and self.chords[-1].midis[0].value == 0:
                    self.chords[-1].quarter_duration += Fraction(self.remaining_duration)
                else:
                    rest = TreeChord(midis=0, quarter_duration=self.remaining_duration)
                    self.add_chord(rest)
            self._filled_with_rest = True

    def preliminary_adjoin_rests(self):
        if not self._filled_with_rest:
            raise Exception('fill_with_rest() first')

        if not self._preliminary_rests_adjoined:
            for ch in self.chords:
                ch.marked = False

            chord_iterator = iter(self.chords)

            def _adjoin(current_chord, next_chord):
                def _chords_are_adjoinable():
                    condition = current_chord.is_adjoinable and next_chord.is_adjoinable
                    return condition

                def _chords_are_rest():
                    condition = current_chord.is_rest and next_chord.is_rest
                    # _print_condition('_chords_are_not_rest', condition)
                    return condition

                if _chords_are_adjoinable() and _chords_are_rest():
                    current_chord.quarter_duration += next_chord.quarter_duration
                    next_chord.marked = True

                    try:
                        next_chord = chord_iterator.__next__()
                        next_chord = _adjoin(current_chord, next_chord)
                    except StopIteration:
                        pass

                return next_chord

            adjoin = True
            current_chord = chord_iterator.__next__()
            try:
                next_chord = chord_iterator.__next__()
            except StopIteration:
                adjoin = False

            while adjoin:
                try:
                    next_chord = _adjoin(current_chord, next_chord)
                    current_chord = next_chord
                    next_chord = chord_iterator.__next__()
                except StopIteration:
                    break

            voice_new_chords = [ch for ch in self.chords if not ch.marked]

            self._chords = voice_new_chords
            self._preliminary_rests_adjoined = True

    def add_beats(self, list_of_beats=None):
        if not self._preliminary_rests_adjoined:
            raise Exception('preliminary_adjoin_rests() first')

        if not self._beats_added:
            self.set_beats(list_of_beats)
            self._add_chords_to_beats()
            self._split_chords_beatwise()
            self._correct_deviations()
            self._beats_added = True
        # else:
        #     warnings.warn('beats already added to {}. No action took place.'.format(self))

    def quantize(self):
        if not self._beats_added:
            raise Exception('add_beats() first')
        if not self._quantized:
            for beat in self.beats:
                beat.quantize()
            self._quantized = True
        # else:
        #     warnings.warn('{} already quantized. No action took place.'.format(self))

    def clear_zero_heads_tails(self):
        for chord in self.chords:
            if chord.quarter_duration == 0 and (chord._head or chord._tail):
                chord.remove_from_score()

    # def adjoin_rests_in_beat(self):
    #     for beat in self.beats:
    #         beat.adjoin_rests()

    def split_not_notatable(self):
        if not self._quantized:
            raise Exception('quantize() first')
        # self._not_notatable_split = False
        if not self._not_notatable_split:
            self._chords = []
            for beat in self.beats:
                beat.split_not_notatable()
                self._chords.extend(beat.chords)
            self._not_notatable_split = True

        # else:
        #     warnings.warn('types of chords in {} already updated. No action took place.'.format(self))

    def implement_flags(self):
        if not self._not_notatable_split:
            raise Exception('split_not_notatable() first')
        if not self._flags_implemented:
            self.remove_chords()
            for beat in self.beats:
                beat.implement_flags()
                self._chords.extend(beat.chords)
            self._flags_implemented = True

    def adjoin_ties(self):
        if not self._flags_implemented:
            raise Exception('implement_flags() first')

        if not self._ties_adjoined:
            notatables = [1, 1.5, 2, 3, 4, 6, 8]
            # notatables = [1, 2, 3, 4, 6, 8]

            # chord_iterator = iter(reversed(self.chords))
            chord_iterator = iter(self.chords)

            def _adjoin(current_chord, next_chord):

                def _chords_are_adjoinable():
                    condition = current_chord.is_adjoinable and next_chord.is_adjoinable
                    return condition

                def _current_chord_is_all_tied():
                    condition = current_chord.is_tied_to_next
                    return condition

                def _chords_are_not_rest():
                    condition = not current_chord.is_rest and not next_chord.is_rest
                    # _print_condition('_chords_are_not_rest', condition)
                    return condition

                def _chords_have_right_positions():
                    # print 'in _chords_have_right_positions', current_chord.name, next_chord.name
                    if current_chord.quarter_duration == 0.5:
                        if current_chord.parent_beat.best_div in [6, 12]:
                            condition = False
                        else:
                            condition = current_chord.offset % (
                                1) in [0, 0.5] and next_chord.offset % 1 == 0
                    else:
                        condition = current_chord.offset % (
                            1) == 0 and next_chord.offset % 1 == 0
                    # _print_condition('_chords_have_right_positions', condition)
                    return condition

                def _result_is_notatable():
                    condition = current_chord.quarter_duration + next_chord.quarter_duration in notatables
                    # _print_condition('_result_is_notatable', condition)
                    return condition

                if _chords_are_adjoinable() and _current_chord_is_all_tied() and _chords_are_not_rest() and _chords_have_right_positions() and _result_is_notatable() and next_chord.parent_beat.best_div != 6:

                    current_chord.quarter_duration += next_chord.quarter_duration

                    if 'stop' in next_chord.tie_types and 'start' not in next_chord.tie_types:
                        current_chord.remove_tie('start')

                    try:
                        notations = next_chord.get_children_by_type(Notations)
                        slurs = []
                        for notation in notations:
                            slurs.extend(notation.get_children_by_type(Slur))
                        for slur in slurs:
                            current_chord.add_slur_object(slur)
                    except IndexError:
                        pass

                    next_chord.marked = True

                    try:
                        next_chord = chord_iterator.__next__()
                        next_chord = _adjoin(current_chord, next_chord)
                    except StopIteration:
                        pass

                return next_chord

            adjoin = True
            current_chord = chord_iterator.__next__()
            try:
                next_chord = chord_iterator.__next__()
            except StopIteration:
                adjoin = False

            while adjoin:
                try:
                    next_chord = _adjoin(current_chord, next_chord)
                    current_chord = next_chord
                    next_chord = chord_iterator.__next__()
                except StopIteration:
                    break

            voice_new_chords = []
            for chord in self.chords:

                try:
                    if chord.marked:
                        chord.parent_beat.chords.remove(chord)
                    else:
                        voice_new_chords.append(chord)
                except AttributeError as err:
                    voice_new_chords.append(chord)

            self._chords = voice_new_chords

            self._ties_adjoined = True
        # else:
        #     warnings.warn('ties of chords in {} already adjoined. No action took place.'.format(self))

    def adjoin_rests(self):
        if not self._ties_adjoined:
            raise Exception('adjoin_ties() first')
        # self._rests_adjoined = True
        if not self._rests_adjoined:
            # notatables = [1, 1.5, 2, 3, 4, 6, 8]
            notatables = [1, 2, 3, 4, 6, 8]

            chord_iterator = iter(self.chords)

            def _adjoin(current_chord, next_chord):

                def _chords_are_adjoinable():
                    condition = current_chord.is_adjoinable and next_chord.is_adjoinable
                    return condition

                def _chords_are_rest():
                    condition = current_chord.is_rest and next_chord.is_rest
                    # _print_condition('_chords_are_not_rest', condition)
                    return condition

                def _chords_have_right_positions():
                    # print 'in _chords_have_right_positions', current_chord.name, next_chord.name
                    condition = current_chord.offset % (
                        1) == 0 and next_chord.offset % 1 == 0
                    # _print_condition('_chords_have_right_positions', condition)
                    return condition

                def _result_is_notatable():
                    condition = current_chord.quarter_duration + next_chord.quarter_duration in notatables
                    # _print_condition('_result_is_notatable', condition)
                    return condition

                if _chords_are_adjoinable() and _chords_are_rest() and _chords_have_right_positions() and _result_is_notatable() and next_chord.parent_beat.best_div != 6:

                    current_chord.quarter_duration += next_chord.quarter_duration

                    next_chord.marked = True

                    try:
                        next_chord = chord_iterator.__next__()
                        next_chord = _adjoin(current_chord, next_chord)
                    except StopIteration:
                        pass

                return next_chord

            adjoin = True
            current_chord = chord_iterator.__next__()
            try:
                next_chord = chord_iterator.__next__()
            except StopIteration:
                adjoin = False

            while adjoin:
                try:
                    next_chord = _adjoin(current_chord, next_chord)
                    current_chord = next_chord
                    next_chord = chord_iterator.__next__()
                except StopIteration:
                    break

            voice_new_chords = []
            for chord in self.chords:
                try:
                    if chord.marked:
                        chord.parent_beat.chords.remove(chord)
                    else:
                        voice_new_chords.append(chord)
                except AttributeError:
                    voice_new_chords.append(chord)

            self._chords = voice_new_chords

            self._rests_adjoined = True
        # else:
        #     warnings.warn('rests in {} already adjoin_rests. No action took place.'.format(self))

    def update_tuplets(self):
        if not self._rests_adjoined:
            raise Exception('adjoin_rests() first')

        if not self._tuplets_updated:
            for beat in self.beats:
                beat.update_tuplets()
            self._tuplets_updated = True
        # else:
        #     warnings.warn('types of chords in {} already updated. No action took place.'.format(self))

    def substitute_sextoles(self):
        if not self._tuplets_updated:
            raise Exception('update_tuplets() first')

        if not self._sextoles_substituted:
            for beat in self.beats:
                beat.substitute_sextoles()
            self._sextoles_substituted = True
        # else:
        #     warnings.warn('all sextoles in  {} already checked. No action took place.'.format(self))

    def update_types(self):
        if not self._sextoles_substituted:
            raise Exception('substitute_sextoles() first')

        if not self._types_updated:
            for chord in self.chords:
                chord.update_type()
            self._types_updated = True
        # else:
        #     warnings.warn('types of chords in {} already updated. No action took place.'.format(self))

    def update_dots(self):
        if not self._types_updated:
            raise Exception('update_types() first')

        if not self._dots_updated:
            for chord in self.chords:
                if chord.quarter_duration != 0:
                    chord.update_dot()
            self._dots_updated = True
        # else:
        #     warnings.warn('types of chords in {} already updated. No action took place.'.format(self))

    def implement_flags_2(self):
        def check_implement_output(chords):
            if not isinstance(chords, list):
                raise Exception('output of implement can only be a list of chords')

            # if len(chords) not in [1, 2]:
            #     raise Exception('output of implement can have 1 or 2 chords')

            for ch in chords:
                if not isinstance(ch, TreeChord):
                    raise Exception('output of implement can only be a list of chords')

        if not self._flags2_implemented:
            flag_types = set([flag.__class__ for flag in flatten([chord.flags for chord in self.chords]) if
                              isinstance(flag, TreeChordFlag2)])

            while flag_types:
                flag_type = flag_types.pop()
                output = []
                for chord in self.chords:
                    try:
                        chord_flag = [flag for flag in chord.flags if isinstance(flag, flag_type)][0]
                        # print(chord)
                        new_chords = chord_flag.implement(chord)
                        check_implement_output(new_chords)
                        output.extend(new_chords)

                    except IndexError:
                        output.append(chord)

                self.remove_chords()
                for chord in output:
                    try:
                        v = chord.get_children_by_type(Voice)[0]
                        chord.remove_child(v)
                    except IndexError:
                        pass
                    if self.add_chord(chord) is not None:
                        raise Exception()
        self._flags2_implemented = True

    def implement_flags_3(self):
        def check_implement_output(chords):
            if not isinstance(chords, list):
                raise Exception('output of implement can only be a list of chords')

            # if len(chords) not in [1, 2]:
            #     raise Exception('output of implement can have 1 or 2 chords')

            for ch in chords:
                if not isinstance(ch, TreeChord):
                    raise Exception('output of implement can only be a list of chords not {}'.format(type(ch)))

        if not self._flags3_implemented:
            flag_types = set([flag.__class__ for flag in flatten([chord.flags for chord in self.chords]) if
                              isinstance(flag, TreeChordFlag3)])

            while flag_types:
                flag_type = flag_types.pop()
                output = []
                for chord in self.chords:
                    try:
                        chord_flag = [flag for flag in chord.flags if isinstance(flag, flag_type)][0]
                        new_chords = chord_flag.implement(chord)
                        check_implement_output(new_chords)
                        output.extend(new_chords)

                    except IndexError:
                        output.append(chord)

                self.remove_chords()
                for chord in output:
                    try:
                        v = chord.get_children_by_type(Voice)[0]
                        chord.remove_child(v)
                    except IndexError:
                        pass
                    tmp = self.add_chord(chord)
                    # if tmp is not None:
                    #     print(tmp.quarter_duration)
                    #     raise Exception()

        self._flags3_implemented = True


class TreePart(timewise.Part):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        attributes = self.add_child(Attributes())
        attributes.add_child(Divisions(1))
        self._max_division = None
        self._forbidden_divisions = None
        self._voices = {}
        self._chords_notated = False
        self._divisions_updated = False
        self._finished = False
        self.parent_score_part = None
        self._accidental_mode = 'normal'

    @property
    def __name__(self):
        index = self.up.get_children_by_type(self.__class__).index(self)
        return self.up.__name__ + '.' + str(index + 1)

    @property
    def chords(self):
        output = []
        for voice in self.voices.values():
            output.extend(voice.chords)
        return output

    @property
    def voices(self):
        return self._voices

    @property
    def max_division(self):
        if self._max_division is None:
            self._max_division = self.parent_score_part.max_division
        return self._max_division

    @max_division.setter
    def max_division(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError('max_division.value must be None or of type int not {}'.format(type(value)))

        self._max_division = value

    @property
    def forbidden_divisions(self):
        if self._forbidden_divisions is None:
            self._forbidden_divisions = self.parent_score_part.forbidden_divisions

        return self._forbidden_divisions

    @forbidden_divisions.setter
    def forbidden_divisions(self, value):
        if value is not None:
            for x in value:
                if not isinstance(x, int):
                    raise TypeError('forbidden_division must be of type int not{}'.format(type(value)))

        self._forbidden_divisions = value

    @property
    def accidental_mode(self):
        return self._accidental_mode

    @accidental_mode.setter
    def accidental_mode(self, value):
        self._accidental_mode = value

    @property
    def xml_chords(self):
        output = []
        for voice in self.voices.values():
            output.extend(voice.xml_chords)
        return output

    def get_divisions(self):
        duration_denominators = [chord.quarter_duration.denominator for chord in
                                 self.chords]

        if len(duration_denominators) == 0:
            return 1
        elif len(duration_denominators) == 1:
            return duration_denominators[0]
        else:
            return lcm(duration_denominators)

    @property
    def notes(self):
        return self.get_children_by_type(TreeNote)

    def set_voice(self, voice_number):
        self.voices[voice_number] = TreePartVoice(voice_number)
        self.voices[voice_number].parent_part = self
        return self.voices[voice_number]

    def get_voice(self, voice_number):
        try:
            return self.voices[voice_number]
        except KeyError:
            return self.set_voice(voice_number)

    def add_chord(self, chord, voice_number=1):
        measure = self.up
        if not measure:
            raise Exception('parent measure needed before adding chord to part')

        if not isinstance(chord, TreeChord):
            raise TypeError()

        voice = self.get_voice(voice_number)
        return voice.add_chord(chord)

    def remove_chord(self, chord):
        if not isinstance(chord, TreeChord):
            raise TypeError()

        tree_voice = self.get_voice(chord.parent_voice.number)
        xml_voice = chord.get_children_by_type(Voice)[0]
        chord.remove_child(xml_voice)
        tree_voice.chords.remove(chord)

    def get_beats(self):
        output = []
        for voice in self.voices.values():
            output.extend(voice.get_beats())
        return output

    def group_beams(self):
        for voice in self.voices.values():
            voice.group_beams()

    def chords_to_notes(self):
        for voice in self.voices.values():
            if not voice._dots_updated:
                raise Exception('update_dots() first')

        if not self.up.previous and self.parent_score_part and self.parent_score_part.instrument:
            clef = self.parent_score_part.instrument.standard_clef
            if clef:
                first_chords = [voice.chords[0] for voice in self.voices.values()]
                first_clefs = [chord.get_clef() for chord in first_chords if chord.get_clef()]
                if not first_clefs:
                    first_chords[0].add_clef(clef)

        if not self._chords_notated:
            for index, voice in enumerate(self.voices.values()):
                if index != 0:
                    self.add_child(TreeBackup(quarter_duration=voice.parent_part.up.quarter_duration))
                for chord in voice.chords:
                    for direction in chord.get_children_by_type(Direction):
                        self.add_child(direction)
                    for attributes in chord.get_children_by_type(Attributes):
                        self.add_child(attributes)
                    for note in chord._notes:
                        self.add_child(note)
            self._chords_notated = True
        else:
            warnings.warn('chord in {} are already notated. No action took place.'.format(self))

    def update_divisions(self):
        if not self._chords_notated:
            raise Exception('chord_to_notes() first')

        if not self._divisions_updated:
            attributes = self.get_children_by_type(Attributes)[0]
            divisions = attributes.get_children_by_type(Divisions)[0]
            divisions.value = self.get_divisions()
            self._divisions_updated = True
        else:
            warnings.warn('divisions in {} are already updated. No action took place.'.format(self))

    def get_previous_measure_last_notes(self):
        previous_measure_last_notes = []
        try:
            previous_measure = self.up.previous
        except AttributeError:
            previous_measure = None
        if previous_measure:
            part = [p for p in previous_measure.get_children_by_type(TreePart) if p.id == self.id][0]
            voices = part.voices.values()
            previous_measure_last_chords = []
            for voice in voices:
                previous_measure_last_chords.append(voice.chords[-1])

            previous_measure_last_notes = []
            for chord in previous_measure_last_chords:
                previous_measure_last_notes.extend(chord._notes)
        return previous_measure_last_notes

    def update_accidentals(self, mode):
        self._accidental_mode = mode

        def _get_previous_measure_last_signed_notes():
            previous_measure_last_notes = self.get_previous_measure_last_notes()
            return [n for n in previous_measure_last_notes if
                    isinstance(n.event, Pitch) and n.pitch.alter and n.pitch.alter.value != 0]

        def get_accidental_info(note):
            try:
                return (note.pitch.step.value, note.pitch.alter.value)
            except AttributeError:
                return (note.pitch.step.value, 0)

        def is_in_repetition(xml_chord):
            if xml_chord.is_tied_to_previous:
                return False

            previous = xml_chord.previous
            while previous:
                if previous.is_tied_to_previous:
                    previous = previous.previous
                else:
                    break

            if previous and xml_chord.has_same_pitches(xml_chord.previous):
                return True

        if mode == 'normal':
            _hide_accidental = []
            _set_natural = set()
            pitched_notes = [note for note in self.get_children_by_type(TreeNote) if isinstance(note.event, Pitch)]
            _first_chord_natural = [note.pitch.step.value for note in
                                    _get_previous_measure_last_signed_notes()]

            for note in pitched_notes:
                if note.pitch.alter and note.pitch.alter.value != 0 and (
                        note.pitch.step.value, note.pitch.alter.value) not in _hide_accidental:
                    if 'stop' not in [t.type for t in note.get_children_by_type(Tie)]:
                        note.accidental.show = True
                        _hide_accidental.append(get_accidental_info(note))
                    _set_natural.add(note.pitch.step.value)
                elif (
                        not note.pitch.alter or note.pitch.alter.value == 0):
                    if note.pitch.step.value in _set_natural:
                        for item in _hide_accidental:
                            if item[0] == note.pitch.step.value:
                                _hide_accidental.remove(item)

                        _set_natural.remove(note.pitch.step.value)
                        note.accidental.show = True
                    elif note.offset == 0:
                        if note.pitch.step.value in _first_chord_natural and 'stop' not in [t.type for t in
                                                                                            note.get_children_by_type(
                                                                                                Tie)]:
                            note.accidental.show = True

        elif mode == 'modern':
            _first_chord_natural = [note.pitch.step.value for note in _get_previous_measure_last_signed_notes()]
            _set_natural = set()
            for index, xml_chord in enumerate(self.xml_chords):
                # first xml_chord
                if index == 0:
                    for note in xml_chord.notes:
                        if isinstance(note.event, Rest):
                            break
                        # natural
                        elif not note.pitch.alter or (note.pitch.alter and note.pitch.alter.value == 0):
                            if note.pitch.step.value in _first_chord_natural and 'stop' not in [t.type for t in
                                                                                                note.get_children_by_type(
                                                                                                    Tie)]:
                                note.accidental.show = True
                        # non natural
                        else:
                            _set_natural.add(note.pitch.step.value)
                            if 'stop' not in [t.type for t in note.get_children_by_type(Tie)]:
                                note.accidental.show = True

                else:
                    # other xml_chords
                    for note in xml_chord.notes:
                        if isinstance(note.event, Rest):
                            break
                        # natural
                        if not note.pitch.alter or (note.pitch.alter and note.pitch.alter.value == 0):
                            if note.pitch.step.value in _set_natural:
                                note.accidental.show = True
                                _set_natural.remove(note.pitch.step.value)
                        # non natural
                        else:
                            if note.is_finger_tremolo:
                                note.accidental.show = True
                            elif is_in_repetition(xml_chord):
                                break
                            elif 'stop' not in [t.type for t in
                                                note.get_children_by_type(Tie)]:
                                note.accidental.show = True

                            _set_natural.add(note.pitch.step.value)
        else:
            raise MusicTreeError('mode {} is not known to update accidentals'.format(mode))

    def fill_with_rest(self):
        if self.voices == {}:
            self.set_voice(1)
        for voice in self.voices.values():
            voice.fill_with_rest()

    def preliminary_adjoin_rests(self):
        for voice in self.voices.values():
            voice.preliminary_adjoin_rests()

    def add_beats(self, list_of_beats=None):
        for voice in self.voices.values():
            voice.add_beats(list_of_beats)

    def quantize(self):
        for voice in self.voices.values():
            voice.quantize()
            voice.clear_zero_heads_tails()

    # def adjoin_rests_in_beat(self):
    #     for voice in self.voices.values():
    #         voice.adjoin_rests_in_beat()

    def split_not_notatable(self):
        for voice in self.voices.values():
            voice.split_not_notatable()

    def implement_flags(self):
        for voice in self.voices.values():
            voice.implement_flags()

    def adjoin_ties(self):
        for voice in self.voices.values():
            voice.adjoin_ties()

    def adjoin_rests(self):
        for voice in self.voices.values():
            voice.adjoin_rests()

    def update_tuplets(self):
        for voice in self.voices.values():
            voice.update_tuplets()

    def substitute_sextoles(self):
        for voice in self.voices.values():
            voice.substitute_sextoles()

    def update_types(self):
        for voice in self.voices.values():
            voice.update_types()

    def update_dots(self):
        for voice in self.voices.values():
            voice.update_dots()

    def implement_flags_2(self):
        for voice in self.voices.values():
            voice.implement_flags_2()

    def implement_flags_3(self):
        for voice in self.voices.values():
            voice.implement_flags_3()

    def update_durations(self):
        for note in self.get_children_by_type(TreeNote):
            if note.quarter_duration != 0:
                note.update_duration(self.get_divisions())
        for backup in self.get_children_by_type(TreeBackup):
            backup.update_duration(self.get_divisions())

    def add_metronome(self, beat_unit='quarter', per_minute=60, **kwargs):
        d = self.add_child(Direction())
        dt = d.add_child(DirectionType())
        m = dt.add_child(Metronome(**kwargs))
        m.add_child(BeatUnit(beat_unit))
        m.add_child(PerMinute(str(per_minute)))
        d.add_child(Sound(tempo=per_minute))

    def add_voice(self, voice):
        self.voices[voice.number] = voice

    def finish(self):
        if not self._finished:
            self.fill_with_rest()

            self.preliminary_adjoin_rests()

            self.add_beats()

            self.quantize()

            # self.adjoin_rests_in_beat()

            self.split_not_notatable()

            self.implement_flags()

            self.adjoin_ties()

            self.adjoin_rests()

            self.update_tuplets()

            self.substitute_sextoles()

            self.implement_flags_2()

            self.update_types()

            self.update_dots()

            self.group_beams()

            self.implement_flags_3()

            self.chords_to_notes()

            self.update_divisions()

            self.update_accidentals(mode=self.accidental_mode)

            self.update_durations()

            self.close_dtd()

            self._finished = True

    def to_string(self):
        self.finish()
        xml = self._to_xml()
        return et.tounicode(xml, pretty_print=True)



from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.position import YPosition
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeStemValue


class ComplexTypeStem(ComplexType, TypeStemValue, YPosition, Color):
    """Stems can be down, up, none, or double. For down and up stems, the position attributes can be used to specify
    stem length. The relative values specify the end of the stem relative to the program default. Default values specify
    an absolute end stem position. Negative values of relative-y that would flip a stem instead of shortening it are
    ignored. A stem element associated with a rest refers to a stemlet."""

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.types.simple_type import TypeBarStyle
from musicscore.musicxml.elements.xml_element import XMLElement
from unittest import TestCase


class Bared(XMLElement, TypeBarStyle):

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag='test_bar', value=value, *args, **kwargs)


class TestBarstyleType(TestCase):
    def test_barstyle(self):
        bst = TypeBarStyle('regular')
        with self.assertRaises(ValueError):
            bst = TypeBarStyle('bla')

class TestBaredXML(TestCase):
    def test_barstyled_xml(self):
        bared = Bared('light-light')
        # bared.value = 'regular'
        with self.assertRaises(ValueError):
            bared.value = 'bla'
        result = '''<test_bar>light-light</test_bar>
'''
        self.assertEqual(bared.to_string(), result)





from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.attributes.trillsound import TrillSound
from musicscore.musicxml.types.complextypes.complextype import ComplexType


class ComplexTypeEmptyTrillSound(ComplexType, PrintStyle, Placement, TrillSound):
    """
    The empty-trill-sound type represents an empty element with print-style, placement, and trill-sound attributes.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)





from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechordflags import BeatwiseFlag
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        durations = [5]
        self.score.set_time_signatures(times={1: (5, 4)})
        sf = SimpleFormat(durations=durations)

        sf.to_stream_voice().add_to_score(self.score, part_number=1)
        sf.chords[0].add_flag(BeatwiseFlag())
        sf.to_stream_voice().add_to_score(self.score, part_number=2)

        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_2(self):
        # todo
        xml_path = path + '_test_2.xml'
        durations = [4, 5, 6, 7, 8, 9, 10]
        self.score.set_time_signatures(
            times={1: (4, 4), 2: (5, 4), 3: (6, 4), 4: (7, 4), 5: (8, 4), 6: (9, 4), 7: (10, 4)})
        sf = SimpleFormat(durations=durations)

        sf.to_stream_voice().add_to_score(self.score, part_number=1)
        for chord in sf.chords:
            chord.add_flag(BeatwiseFlag())
        sf.to_stream_voice().add_to_score(self.score, part_number=2)

        self.score.write(xml_path)
        # TestScore().assert_template(xml_path)

    def test_3(self):
        xml_path = path + '_test_3.xml'
        durations = [0.5, 1.5]
        self.score.set_time_signatures(durations=2)
        sf = SimpleFormat(durations=durations)
        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_4(self):
        xml_path = path + '_test_4.xml'
        durations = [0.5, 1.5]
        self.score.set_time_signatures(durations=2)
        sf = SimpleFormat(durations=durations, midis=[60, 0])
        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_5(self):
        xml_path = path + '_test_5.xml'
        durations = [0.5, 3.5]
        sf = SimpleFormat(durations=durations, midis=[60, 0])
        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)

    def test_6(self):
        xml_path = path + '_test_6.xml'
        durations = [0.5, 1.5, 2]
        sf = SimpleFormat(durations=durations, midis=[60, 0, 60])
        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)



from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.symbolformatting import SymbolFormatting
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeSmulfGlyphName


class ComplexTypeFormattedSymbolId(ComplexType, TypeSmulfGlyphName, SymbolFormatting, OptionalUniqueId):
    """
    The formatted-symbol-id type represents a SMuFL musical symbol element with formatting and id attributes.

    """

    def __init__(self, tag, value=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)



"""
    <xs:complexType name="supports">
        <xs:attribute name="attribute" type="xs:NMTOKEN"/>
        <xs:attribute name="value" type="xs:token"/>
    </xs:complexType>
"""
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeYesNo, Token


class ComplexTypeSupports(ComplexType):
    """
    The supports type indicates if a MusicXML encoding supports a particular MusicXML element. This is recommended for
    elements like beam, stem, and accidental, where the absence of an element is ambiguous if you do not know if the
    encoding supports that element. For Version 2.0, the supports element is expanded to allow programs to indicate
    support for particular attributes or particular values. This lets applications communicate, for example, that all
    system and/or page breaks are contained in the MusicXML file.
    """

    def __init__(self, tag, type_, element, attribute=None, value_=None, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self._ATTRIBUTES = ['attribute', 'element', 'type', 'value']
        self.type_ = type_
        self.element = element
        self.attribute = attribute
        self.value_ = value_

    @property
    def type_(self):
        return self.get_attribute('type')

    @type_.setter
    def type_(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeYesNo(value)
            self.set_attribute('type', value)

    @property
    def attribute(self):
        return self.get_attribute('attribute')

    @attribute.setter
    def attribute(self, value):
        if value is None:
            self.remove_attribute('attribute')
        else:
            Token(value)
            self.set_attribute('attribute', value)

    @property
    def element(self):
        return self.get_attribute('element')

    @element.setter
    def element(self, value):
        if value is None:
            self.remove_attribute('element')
        else:
            Token(value)
            self.set_attribute('element', value)

    @property
    def value_(self):
        return self.get_attribute('value')

    @value_.setter
    def value_(self, v):
        if v is None:
            self.remove_attribute('value')
        else:
            Token(v)
            self.set_attribute('value', v)



from musicscore.musicxml.elements.note import Notehead, Type, Stem, Beam


class TreeChordFlag3(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __deepcopy__(self, memodict={}):
        return self.__class__()


class NoiseFlag3(TreeChordFlag3):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def implement(self, chord):
        chord.add_child(Notehead('square'))
        return [chord]


class BowPositionFlag(TreeChordFlag3):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def implement(self, chord):
        if chord.is_tied_to_previous:
            chord.remove_previous_tie()
            chord.add_child(Notehead('none'))
            chord.add_child(Stem('none'))
        elif chord.is_tied_to_next:
            chord.add_child(Stem('none'))

        else:
            chord.add_child(Stem('none'))
        beams = chord.get_children_by_type(Beam)
        for beam in beams:
            chord.remove_child(beam)
        return [chord]


class CheckPreviousFlag(TreeChordFlag3):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def implement(self, chord):
        # print('CheckPrevious:')
        # print(chord.previous_in_score_part)
        return [chord]



from musicscore.dtd.dtd import Sequence, GroupReference, Element, Choice
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printobject import PrintObject
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.groups.common import Editorial
from musicscore.musicxml.types.complextypes.clef import ComplexTypeClef
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.key import TypeKey
from musicscore.musicxml.types.simple_type import TypePositiveDivisions, PositiveInteger


class Divisions(XMLElement, TypePositiveDivisions):
    """
    Musical notation duration is commonly represented as fractions.
    The divisions element indicates how many divisions per quarter note are used to indicate a note's duration.
    For example, if duration = 1 and divisions = 2, this is an eighth note duration.
    Duration and divisions are used directly for generating sound output, so they must be chosen to take tuplets into
    account. Using a divisions element lets us use just one number to represent a duration for each note in the score,
    while retaining the full power of a fractional representation.
    If maximum compatibility with Standard MIDI 1.0 files is important, do not have the divisions value exceed 16383.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='divisions', value=value, *args, **kwargs)


class Key(TypeKey):
    """
    The key element represents a key signature. Both traditional and non-traditional key signatures are supported.
    The optional number attribute refers to staff numbers. If absent, the key signature applies to all staves in the
    part.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='key', value=value, *args, **kwargs)


class Beats(XMLElement, PositiveInteger):
    """
    The beats element indicates the number of beats, as found in the numerator of a time signature.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='beats', value=value, *args, **kwargs)

    def __copy__(self):
        new_beats = Beats(self.value)
        return new_beats


class BeatType(XMLElement, PositiveInteger):
    """
    The beat-type element indicates the beat unit, as found in the denominator of a time signature.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='beat-type', value=value, *args, **kwargs)

    def __copy__(self):
        new_beat_type = BeatType(self.value)
        return new_beat_type


TimeSignature = Sequence(
    Element(Beats),
    Element(BeatType)
)


class Interchangeable(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='interchangeable', *args, **kwargs)
        raise NotImplementedError()


class SenzaMisura(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='senza-misura', *args, **kwargs)
        raise NotImplementedError()


# todo:	<xs:attribute name="number" type="staff-number"/>
# 		<xs:attribute name="symbol" type="time-symbol"/>
# 		<xs:attribute name="separator" type="time-separator"/>
# 		<xs:attributeGroup ref="print-style-align"/>

# , OptionalUniqueId
class Time(XMLElement, PrintObject, OptionalUniqueId):
    """
    Time signatures are represented by the beats element for the numerator and the beat-type element for the
    denominator. The symbol attribute is used indicate common and cut time symbols as well as a single number display.
    Multiple pairs of beat and beat-type elements are used for composite time signatures with multiple denominators,
    such as 2/4 + 3/8. A composite such as 3+2/8 requires only one beat/beat-type pair.
    The print-object attribute allows a time signature to be specified but not printed, as is the case for excerpts
    from the middle of a score. The value is "yes" if not present. The optional number attribute refers to staff numbers
    within the part. If absent, the time signature applies to all staves in the part.
    """

    _DTD = Choice(
        Sequence(
            GroupReference(TimeSignature, max_occurrence=None),
            Element(Interchangeable, min_occurrence=0)
        ),
        Element(SenzaMisura)
    )

    def __init__(self, *args, **kwargs):
        super().__init__(tag='time', *args, **kwargs)


class Staves(XMLElement):
    """
    The staves element is used if there is more than one staff represented in the given part (e.g., 2 staves for typical
    piano parts). If absent, a value of 1 is assumed. Staves are ordered from top to bottom in a part in numerical
    order, with staff 1 above staff 2

        </xs:element>
    <xs:element name="staves" type="xs:nonNegativeInteger" minOccurs="0">
    </xs:element>

    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='staves', value=value, *args, **kwargs)
        raise NotImplementedError()


class PartSymbol(XMLElement):
    """
    The part-symbol element indicates how a symbol for a multi-staff part is indicated in the score.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='part-symbols', value=value, *args, **kwargs)
        raise NotImplementedError()


class Instruments(XMLElement):
    """
    <xs:element name="instruments" type="xs:nonNegativeInteger" minOccurs="0">
      <xs:annotation>
        <xs:documentation>The instruments element is only used if more than one instrument is represented in the part (e.g., oboe I and II where they play together most of the time). If absent, a value of 1 is assumed.</xs:documentation>
      </xs:annotation>
    </xs:element>
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='instruments', value=value, *args, **kwargs)
        raise NotImplementedError()


class Clef(ComplexTypeClef):
    """
    Clefs are represented by a combination of sign, line, and clef-octave-change elements.
    The optional number attribute refers to staff numbers within the part.
    A value of 1 is assumed if not present.
    Sometimes clefs are added to the staff in non-standard line positions, either to indicate cue passages,
    or when there are multiple clefs present simultaneously on one staff. In this situation,
    the additional attribute is set to "yes" and the line value is ignored.
    The size attribute is used for clefs where the additional attribute is "yes".
    It is typically used to indicate cue clefs.

    Sometimes clefs at the start of a measure need to appear after the barline rather than before,
    as for cues or for use after a repeated section.
    The after-barline attribute is set to "yes" in this situation. The attribute is ignored for mid-measure clefs.

    Clefs appear at the start of each system unless the print-object attribute has been set to "no" or
    the additional attribute has been set to "yes"
    """
    _TAG = 'clef'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class StaffDetails(XMLElement):
    """
    <xs:element name="staff-details" type="staff-details" minOccurs="0" maxOccurs="unbounded">
      <xs:annotation>
        <xs:documentation>The staff-details element is used to indicate different types of staves.</xs:documentation>
      </xs:annotation>
    </xs:element>
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='staff-details', value=value, *args, **kwargs)
        raise NotImplementedError()


class Transpose(XMLElement):
    """
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='transpose', value=value, *args, **kwargs)
        raise NotImplementedError()


class Directive(XMLElement):
    """
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='directive', value=value, *args, **kwargs)
        raise NotImplementedError()


class MeasureStyle(XMLElement):
    """
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='measure-style', value=value, *args, **kwargs)
        raise NotImplementedError()


class ComplexTypeAttributes(ComplexType):
    """
    The attributes element contains musical information that typically changes on measure boundaries. This includes key
    and time signatures, clefs, transpositions, and staving. When attributes are changed mid-measure, it affects the
    music in score order, not in MusicXML document order
    """
    _DTD = Sequence(
        GroupReference(Editorial),
        Element(Divisions, min_occurrence=0),
        Element(Key, min_occurrence=0, max_occurrence=None),
        Element(Time, min_occurrence=0, max_occurrence=None),
        Element(Staves, min_occurrence=0),
        Element(PartSymbol, min_occurrence=0),
        Element(Instruments, min_occurrence=0),
        Element(Clef, min_occurrence=0),
        Element(StaffDetails, min_occurrence=0, max_occurrence=None),
        Element(Transpose, min_occurrence=0, max_occurrence=None),
        Element(Directive, min_occurrence=0, max_occurrence=None),
        Element(MeasureStyle, min_occurrence=0, max_occurrence=None)

    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()
        sf = SimpleFormat(durations=[1, 1])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        sf = SimpleFormat(durations=[1, 1])
        v = sf.to_stream_voice(2)
        v.add_to_score(self.score, 1, 1)

        sf = SimpleFormat(durations=[1, 1])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 2, 2)

    def test_1(self):
        chord = self.score.get_measure(1).get_part(1).get_voice(1).chords[1]
        self.assertEqual(chord.__name__, '1.1.1.2')
        chord = self.score.get_measure(1).get_part(1).get_voice(2).chords[0]
        self.assertEqual(chord.__name__, '1.1.2.1')
        chord = self.score.get_measure(2).get_part(2).get_voice(1).chords[1]
        self.assertEqual(chord.__name__, '2.2.1.2')



from musicscore.musictree.exceptions import MusicTreeError
from musicscore.musicxml.types.complextypes.attributes import Time, SenzaMisura, Beats, BeatType


class TreeTime(Time):

    def __init__(self, *time_signatures, **kwargs):
        super().__init__(**kwargs)
        self._force_show = False
        self._force_hide = False
        self.pars_arguments(time_signatures)

    @property
    def force_show(self):
        return self._force_show

    @force_show.setter
    def force_show(self, value):
        if not isinstance(value, bool):
            raise TypeError('force_show.value must be of type bool not{}'.format(type(value)))

        self._force_show = value

    @property
    def force_hide(self):
        return self._force_hide

    @force_hide.setter
    def force_hide(self, value):
        if not isinstance(value, bool):
            raise TypeError('force_hide.value must be of type bool not{}'.format(type(value)))
        self._force_hide = value

    @property
    def values(self):
        return [child.value for child in self.get_children() if isinstance(child, Beats) or isinstance(child, BeatType)]

    def pars_arguments(self, time_signatures):
        if len(time_signatures) == 1 and time_signatures[0] == 'senza_misura':
            self.add_child(SenzaMisura())

        elif len(time_signatures) % 2 == 0:

            for time_signature in zip(time_signatures[0::2], time_signatures[1::2]):
                self.set_time_signature(time_signature)
        else:
            raise MusicTreeError(
                'TreeTime can have senza_misura or (beats, beat_type)* as arguments not {}'.format(time_signatures))

    def set_time_signature(self, time_signature):
        (beats, beat_type) = time_signature
        self.add_child(Beats(beats))
        permitted = (1, 2, 4, 8, 16, 32, 64)
        if beat_type not in permitted:
            raise MusicTreeError('beat_type {} must be in {}'.format(beats, permitted))
        else:
            self.add_child(BeatType(beat_type))

    def get_time_signatures(self):
        if self.get_children_by_type(SenzaMisura):
            return []
        else:
            return list(zip(self.get_children_by_type(Beats), self.get_children_by_type(BeatType)))

    def __copy__(self):
        new_time = TreeTime()
        for key, new_key in zip(self.__dict__.keys(), new_time.__dict__.keys()):
            item = self.__dict__[key]
            if key == '_attributes':
                new_time.__dict__[new_key] = item
        for xml_child in self.get_children():
            new_time.add_child(xml_child.__copy__())

        return new_time



from quicktions import Fraction
from unittest import TestCase
import musicscore.musictree.treebeat as q
from musicscore.musictree.treechord import TreeChord


class Test(TestCase):

    def test_1(self):
        durations = [0.4, 0.2, 0.1, 0.3]
        b = q.TreeBeat(max_division=5, forbidden_divisions=[])

        result = [Fraction(2, 5), Fraction(1, 5), Fraction(1, 5), Fraction(1, 5)]

        self.assertEqual(b.get_quantized_durations(durations), result)
        self.assertEqual(b.best_div, 5)

    def test_2(self):
        b = q.TreeBeat(max_division=5, forbidden_divisions=[])
        b.add_chord(TreeChord((60, 61), quarter_duration=1))
        b.quantize()
        self.assertEqual([chord.quarter_duration for chord in b.chords], [1])

    def test_3(self):
        b = q.TreeBeat(max_division=5, forbidden_divisions=[])
        b.add_chord(TreeChord((60, 61), quarter_duration=0.5))
        b.add_chord(TreeChord((60, 61), quarter_duration=0.5))
        b.quantize()
        self.assertEqual([chord.quarter_duration for chord in b.chords], [Fraction(1, 2), Fraction(1, 2)])

    def test_4(self):
        b = q.TreeBeat(max_division=5, forbidden_divisions=[])
        b.add_chord(TreeChord((60, 61), 5))
        b.quantize()
        self.assertEqual([chord.quarter_duration for chord in b.chords], [5])

    def test_5(self):
        b = q.TreeBeat(max_division=7, forbidden_divisions=[])
        b.add_chord(TreeChord((60, 61), quarter_duration=0.5))
        b.add_chord(TreeChord((60, 61), quarter_duration=0.2))
        b.add_chord(TreeChord((60, 61), quarter_duration=0.3))
        b.quantize()
        self.assertEqual([chord.quarter_duration for chord in b.chords],
                         [Fraction(1, 2), Fraction(1, 6), Fraction(1, 3)])


from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeLineWidthType, TypeTenths


class ComplexTypeLineWidth(ComplexType, TypeTenths):
    """The line-width type indicates the width of a line type in tenths. The type attribute defines what type of line
    is being defined. Values include beam, bracket, dashes, enclosure, ending, extend, heavy barline, leger, light
    barline, octave shift, pedal, slur middle, slur tip, staff, stem, tie middle, tie tip, tuplet bracket, and wedge.
    The text content is expressed in tenths."""

    _ATTRIBUTES = []

    def __init__(self, tag, type, value=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)
        self.type = type

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeLineWidthType(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)



from unittest import TestCase

from musicscore.musicxml.types.complextypes.attributes import Time, Beats, BeatType


class Test(TestCase):

    def test_dtd_time(self):
        t = Time()
        t2 = Time()
        self.assertTrue(id(t.dtd) == id(t2.dtd))
        self.assertTrue([id(node) for node in t.dtd.get_choices()] == [id(node) for node in t2.dtd.get_choices()])
        self.assertFalse(id(t.current_dtd_choice) == id(t2.current_dtd_choice))
        t2.add_child(Beats(6))
        t2.add_child(BeatType(2))
        result = '''<time>
  <beats>6</beats>
  <beat-type>2</beat-type>
</time>
'''
        self.assertEqual(t2.to_string(), result)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class LineHeight(AttributeAbstract):
    """
    The line-height attribute specifies text leading. Values are either "normal" or a number representing the percentage
    of the current font height to use for leading. The default is "normal". The exact normal value is implementation-
    dependent, but values between 100 and 120 are recommended.
    """

    def __init__(self, line_height=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('line-height', line_height, "TypeNumberOrNormal")



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treeclef import BASS_CLEF, TREBLE_CLEF, ALTO_CLEF, TENOR_CLEF, PERCUSSION_CLEF, \
    HIGH_TREBLE_CLEF
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(durations=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
        sf.chords[0].add_clef(BASS_CLEF)
        sf.chords[1].add_clef(TREBLE_CLEF)
        sf.chords[2].add_clef(PERCUSSION_CLEF)
        sf.chords[3].add_clef(ALTO_CLEF)
        sf.chords[4].add_clef(TENOR_CLEF)
        sf.chords[6].add_clef(TENOR_CLEF)
        sf.chords[7].add_clef(HIGH_TREBLE_CLEF)

        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.mididevice import ComplexTypeMidiDevice
from musicscore.musicxml.types.complextypes.midiinstrument import ComplexTypeMidiInstrument
from musicscore.musicxml.types.complextypes.offset import ComplexTypeOffset
from musicscore.musicxml.types.complextypes.play import ComplexTypePlay


class Tempo(AttributeAbstract):
    def __init__(self, tempo=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('tempo', tempo, 'TypeNonNegativeDecimal')


class Dynamics(AttributeAbstract):
    def __init__(self, dynamics=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('dynamics', dynamics, 'TypeNonNegativeDecimal')


class Dacapo(AttributeAbstract):
    def __init__(self, dacapo=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('dacapo', dacapo, 'TypeYesNo')


class Segno(AttributeAbstract):
    def __init__(self, segno=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('segno', segno, 'Token')


class Dalsegno(AttributeAbstract):
    def __init__(self, dalsegno=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('dalsegno', dalsegno, 'Token')


class Coda(AttributeAbstract):
    def __init__(self, coda=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('coda', coda, 'Token')


class Tocoda(AttributeAbstract):
    def __init__(self, tocoda=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('tocoda', tocoda, 'Token')


class Divisions(AttributeAbstract):
    def __init__(self, divisions=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('divisions', divisions, 'TypeDivisions')


class ForwardRepeat(AttributeAbstract):
    def __init__(self, forward_repeat=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('forward-repeat', forward_repeat, 'TypeYesNo')


class Fine(AttributeAbstract):
    def __init__(self, fine=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('fine', fine, 'Token')


class TimeOnly(AttributeAbstract):
    def __init__(self, time_only=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('time-only', time_only, 'TypeTimeOnly')


class Pizzicato(AttributeAbstract):
    def __init__(self, pizzicato=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('pizzicato', pizzicato, 'TypeYesNo')


class Pan(AttributeAbstract):
    def __init__(self, pan=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('pan', pan, 'TypeRotationDegrees')


class Elevation(AttributeAbstract):
    def __init__(self, elevation=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('elevation', elevation, 'TypeRotationDegrees')


class DamperPedal(AttributeAbstract):
    def __init__(self, damper_pedal=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('damper-pedal', damper_pedal, 'TypeYesNoNumber')


class SoftPedal(AttributeAbstract):
    def __init__(self, soft_pedal=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('soft-pedal', soft_pedal, 'TypeYesNoNumber')


class SostenutoPedal(AttributeAbstract):
    def __init__(self, sostenuto_pedal=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('sostenuto-pedal-pedal', sostenuto_pedal, 'TypeYesNoNumber')


class MidiDevice(ComplexTypeMidiDevice):
    _TAG = 'midi-device'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class MidiInstrument(ComplexTypeMidiInstrument):
    _TAG = 'midi-instrument'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Play(ComplexTypePlay):
    _TAG = 'play'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Offset(ComplexTypeOffset):
    _TAG = 'offset'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeSound(ComplexType, Tempo, Dynamics, Dacapo, Segno, Dalsegno, Coda, Tocoda, Divisions, ForwardRepeat,
                       Fine, TimeOnly, Pizzicato, Pan, Elevation, DamperPedal, SoftPedal, SostenutoPedal,
                       OptionalUniqueId):
    """The sound element contains general playback parameters. They can stand alone within a part/measure, or be a
    component element within a direction.

    Tempo is expressed in quarter notes per minute. If 0, the sound-generating program should prompt the user at the
    time of compiling a sound (MIDI) file.

    Dynamics (or MIDI velocity) are expressed as a percentage of the default forte value (90 for MIDI 1.0).

    Dacapo indicates to go back to the beginning of the movement. When used it always has the value "yes".

    Segno and dalsegno are used for backwards jumps to a segno sign; coda and tocoda are used for forward jumps to a
    coda sign. If there are multiple jumps, the value of these parameters can be used to name and distinguish them. If
    segno or coda is used, the divisions attribute can also be used to indicate the number of divisions per quarter
    note. Otherwise sound and MIDI generating programs may have to recompute this.

    By default, a dalsegno or dacapo attribute indicates that the jump should occur the first time through, while a
    tocoda attribute indicates the jump should occur the second time through. The time that jumps occur can be changed
    by using the time-only attribute.

    Forward-repeat is used when a forward repeat sign is implied, and usually follows a bar line. When used it always
    has the value of "yes".

    The fine attribute follows the final note or rest in a movement with a da capo or dal segno direction. If numeric,
    the value represents the actual duration of the final note or rest, which can be ambiguous in written notation and
    different among parts and voices. The value may also be "yes" to indicate no change to the final duration.

    If the sound element applies only particular times through a repeat, the time-only attribute indicates which times
    to apply the sound element.

    Pizzicato in a sound element effects all following notes. Yes indicates pizzicato, no indicates arco.

    The pan and elevation attributes are deprecated in Version 2.0. The pan and elevation elements in the midi-
    instrument element should be used instead. The meaning of the pan and elevation attributes is the same as for the
    pan and elevation elements. If both are present, the mid-instrument elements take priority.

    The damper-pedal, soft-pedal, and sostenuto-pedal attributes effect playback of the three common piano pedals and
    their MIDI controller equivalents. The yes value indicates the pedal is depressed; no indicates the pedal is
    released. A numeric value from 0 to 100 may also be used for half pedaling. This value is the percentage that the
    pedal is depressed. A value of 0 is equivalent to no, and a value of 100 is equivalent to yes.

    MIDI devices, MIDI instruments, and playback techniques are changed using the midi-device, midi-instrument, and play
    elements. When there are multiple instances of these elements, they should be grouped together by instrument using
    the id attribute values.

    The offset element is used to indicate that the sound takes place offset from the current score position. If the
    sound element is a child of a direction element, the sound offset element overrides the direction offset element if
    both elements are present. Note that the offset reflects the intended musical position for the change in sound. It
    should not be used to compensate for latency issues in particular hardware configurations."""

    _DTD = Sequence(
        Sequence(
            Element(MidiDevice, min_occurrence=0),
            Element(MidiInstrument, min_occurrence=0),
            Element(Play, min_occurrence=0),
            min_occurrence=0,
            max_occurrence=None
        ),
        Element(Offset, min_occurrence=0)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, * args, **kwargs)



import os
from unittest import TestCase

from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.groups.common import Voice

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        chord = TreeChord()
        chord.add_child(Voice('1'))
        chord.remove_voice()
        self.assertEqual(chord.get_children(), [])



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(durations=[2, 2], midis=[60, 60])
        sf.chords[0].tie_orientation = 'over'
        sf.chords[0].add_tie('start')
        sf.chords[0].is_adjoinable = False
        sf.chords[1].add_tie('stop')

        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_2(self):
        sf = SimpleFormat(durations=[6], midis=[60])
        sf.chords[0].tie_orientation = 'over'

        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_2.xml'
        self.score.write(xml_path)



import itertools
import os
import random
from unittest import TestCase

from quicktions import Fraction

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(durations=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1])
        for index, chord in enumerate(sf.chords):
            chord.add_lyric(index + 1)
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 2)

        sf = SimpleFormat(durations=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        self.score.fill_with_rest()
        self.score.preliminary_adjoin_rests()
        self.score.add_beats()

        for measure in self.score.get_children_by_type(TreeMeasure):
            part = measure.get_part(2)
            for beat in part.get_beats():
                beat.max_division = 7

        result_path = path + '_test_1'
        # with self.assertWarns(UserWarning):
        #     self.score.write(path=result_path)
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        self.score.add_measure(TreeMeasure(time=(3, 4)))
        sf = SimpleFormat(durations=[0.5, 0.6, 0.7, 0.8])
        for index, chord in enumerate(sf.chords):
            chord.add_lyric(index + 1)
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        sf = SimpleFormat(durations=[0.5, 0.6, 0.7, 0.8])
        for index, chord in enumerate(sf.chords):
            chord.add_lyric(index + 1)
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 2)

        self.score.fill_with_rest()
        self.score.preliminary_adjoin_rests()
        self.score.add_beats()

        for measure in self.score.get_children_by_type(TreeMeasure):
            part = measure.get_part(2)
            for beat in part.get_beats():
                beat.max_division = 7

        self.score.quantize()

        result_path = path + '_test_2'
        # with self.assertWarns(UserWarning):
        #     self.score.write(path=result_path)
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_3(self):

        random.seed(3)
        durations = []
        while sum(durations) <= 16:
            duration = random.randrange(0, 2) + (random.random() / 2.)
            durations.append(Fraction(duration).limit_denominator(100))

        def add_to_score(part=1):
            sf = SimpleFormat(durations=durations)
            dynamics = itertools.cycle(['pppp', 'ppp', 'pp', 'p', 'mp', 'mf', 'f', 'ff', 'fff'])

            for index, chord in enumerate(sf.chords):
                chord.add_lyric(index + 1)
                d = chord.add_dynamics(dynamics.__next__())
                d.relative_y = -20
                d.halign = 'center'
            v = sf.to_stream_voice(1)
            v.add_to_score(self.score, 1, part)

        add_to_score(1)
        add_to_score(2)
        add_to_score(3)
        add_to_score(4)
        add_to_score(5)
        add_to_score(6)
        add_to_score(7)
        add_to_score(8)

        self.score.get_score_parts()[0].max_division = 8
        self.score.get_score_parts()[1].max_division = 7
        self.score.get_score_parts()[2].max_division = 6
        self.score.get_score_parts()[3].max_division = 5
        self.score.get_score_parts()[4].max_division = 4
        self.score.get_score_parts()[5].max_division = 3
        self.score.get_score_parts()[6].max_division = 2
        self.score.get_score_parts()[7].max_division = 1
        result_path = path + '_test_3'

        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_4(self):
        sf = SimpleFormat(
            durations=[Fraction(3, 10), Fraction(3, 10), Fraction(3, 10), Fraction(3, 10), Fraction(3, 10),
                       Fraction(3, 2), Fraction(1, 2), Fraction(1, 3)])
        xml_path = path + '_test_4.xml'
        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        TestScore().assert_template(result_path=xml_path)

    def test_5(self):
        self.score.set_time_signatures(
            [Fraction(3, 2)])
        sf = SimpleFormat(durations=[0.666, 0.333, 0.5])
        xml_path = path + '_test_5.xml'
        sf.to_stream_voice().add_to_score(self.score)
        self.score.get_score_parts()[0].max_division = 1
        self.score.write(xml_path)
        # TestScore().assert_template(result_path=xml_path)

    # def test_4(self):
    #
    #     random.seed(1)
    #     durations = []
    #     while sum(durations) <= 16:
    #         duration = random.randrange(0, 2) + (random.random() / 2.)
    #         durations.append(Fraction(duration).limit_denominator(100))
    #
    #     def add_to_score(part=1):
    #         sf = SimpleFormat(durations=durations)
    #         for index, chord in enumerate(sf.chords):
    #             chord.add_lyric(index + 1)
    #         v = sf.to_voice(1)
    #         v.add_to_score(self.score, 1, part)
    #
    #     add_to_score(1)
    #
    #     def print_last_chord_ties():
    #         chord = self.score.get_measure(3).get_part(1).get_voice(1).get_beats()[-1].chords[-1]
    #         try:
    #             print(chord.get_children_by_type(Lyric)[0].get_children_by_type(Text)[0].value)
    #         except:
    #             pass
    #         print(chord.tie_types)
    #
    #     self.score.get_score_parts()[0].max_division = 2
    #     # self.score.fill_with_rest()
    #     # self.score.add_beats()
    #     # print_last_chord_ties()
    #     # self.score.quantize()
    #     # print_last_chord_ties()
    #     # self.score.split_not_notatable()
    #     # print_last_chord_ties()
    #     # self.score.update_tuplets()
    #     # self.score.substitute_sextoles()
    #     # self.score.update_types()
    #     # self.score.update_dots()
    #     # self.score.group_beams()
    #     # print_last_chord_ties()
    #     # self.score.chord_to_notes()
    #     # print(self.score.get_measure(3).get_part(1).notes[-1].get_children())
    #     # self.score.update_divisions()
    #     #
    #     # self.score.update_accidentals(mode='normal')
    #     #
    #     # self.score.update_durations()
    #     # print(self.score.get_measure(3).get_part(1).notes[-1].get_children())
    #     # self.score.close_dtd()
    #     # print(self.score.get_measure(3).get_part(1).notes[-1].get_children())
    #
    #     result_path = path + '_test_4'
    #     self.score.write(path=result_path)



from unittest import TestCase

from musicscore.musicxml.elements.note import Notations
from musicscore.musicxml.types.complextypes.notations import Fermata


class Test(TestCase):
    def test_1(self):
        fermata = Fermata()
        notations = Notations()
        notations.add_child(fermata)
        # print(notations.to_string())
        result = '''<notations>
  <fermata>normal</fermata>
</notations>
'''
        self.assertEqual(notations.to_string(), result)



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(durations=[0.5, 1.5, 0.3, 1.7])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_1'
        self.score.fill_with_rest()
        self.score.preliminary_adjoin_rests()
        self.score.add_beats()
        self.score.quantize()
        self.score.split_not_notatable()
        for chord in self.score.get_measure(1).get_part(1).get_voice(1).chords:
            chord.add_lyric(round(float(chord.offset), 2))
        # with self.assertWarns(UserWarning):
        #     self.score.write(path=result_path)
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from unittest import TestCase

from musicscore.musicxml.groups.common import Voice
from musicscore.musicxml.elements.fullnote import Pitch
from musicscore.musicxml.elements.note import Note, Duration


class Test(TestCase):
    def setUp(self) -> None:
        self.note = Note()
        self.note.add_child(Pitch())
        self.note.add_child(Duration())

    def test_voice(self):
        self.note.add_child(Voice('1'))
        result = '''<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <voice>1</voice>
</note>
'''
        self.assertEqual(self.note.to_string(), result)


from musicscore.musicxml.elements.scoreheader import Defaults
from musicscore.musicxml.groups.layout import PageLayout, StaffLayout, SystemLayout
from musicscore.musicxml.groups.margins import LeftMargin, RightMargin, TopMargin, BottomMargin
from musicscore.musicxml.types.complextypes.defaults import Scaling
from musicscore.musicxml.types.complextypes.encoding import Supports
from musicscore.musicxml.types.complextypes.identification import Encoding
from musicscore.musicxml.types.complextypes.pagelayout import PageMargins, PageHeight, PageWidth
from musicscore.musicxml.types.complextypes.scaling import Millimeters, Tenths
from musicscore.musicxml.types.complextypes.scorepart import Identification
from musicscore.musicxml.types.complextypes.stafflayout import StaffDistance
from musicscore.musicxml.types.complextypes.systemlayout import SystemDistance, SystemMargins, TopSystemDistance


class TreePageStyle(object):
    sizes = {'A4': (210, 297), 'A3': (297, 420)}

    def __init__(self, score, scale=1, size='A4', orientation='portrait', left_margin=111, right_margin=55, top_margin=83,
                 bottom_margin=83, system_distance=None, system_left_margin=None, system_right_margin=None,
                 top_system_distance=None, staff_distance=None):

        self._score = None
        self.score = score
        self.tenth = 40

        self._size = None
        self._orientation = 'portrait'

        self._defaults = None
        self._scaling = None

        self._page_layout = None
        self._page_height = None
        self._page_width = None
        self._page_margins = None

        self._left_margin = None
        self._right_margin = None
        self._top_margin = None
        self._bottom_margin = None

        self._system_layout = None
        self._system_margins = None
        self._system_distance = None
        self._staff_layout = None
        self._staff_distance = None
        self._system_left_margin = None
        self._system_right_margin = None
        self._top_system_distance = None

        self.scale = scale
        self.size = size
        self.orientation = orientation

        self.left_margin = left_margin
        self.right_margin = right_margin
        self.top_margin = top_margin
        self.bottom_margin = bottom_margin
        self.staff_distance = staff_distance

        self.system_distance = system_distance
        self.system_left_margin = system_left_margin
        self.system_right_margin = system_right_margin
        self.top_system_distance = top_system_distance

    def millimeters_to_tenth(self, m):
        return round(m / self.millimeters * self.tenth)

    @property
    def millimeters(self):
        return self.scale * 7.2319

    def _add_defaults(self):
        self._defaults = self.score.add_child(Defaults())

    def _add_scaling(self):
        if not self._defaults:
            self._add_defaults()
        self._scaling = self._defaults.add_child(Scaling())
        self._scaling.add_child(Millimeters(self.millimeters))
        self._scaling.add_child(Tenths(self.tenth))

    def _add_page_layout(self):
        if not self._scaling:
            self._add_scaling()

        self._page_layout = self._defaults.add_child(PageLayout())

    def _add_staff_layout(self):
        if not self._scaling:
            self._add_scaling()

        self._staff_layout = self._defaults.add_child(StaffLayout())

    def _add_system_layout(self):
        if not self._scaling:
            self._add_scaling()

        self._system_layout = self._defaults.add_child(SystemLayout())
        self._system_distance = self._system_layout.add_child(SystemDistance(100))
        self._system_margins = self._system_layout.add_child(SystemMargins())
        self._system_left_margin = self._system_margins.add_child(LeftMargin(0))
        self._system_right_margin = self._system_margins.add_child(RightMargin(0))
        self._top_system_distance = self._system_layout.add_child(TopSystemDistance(150))

    def _add_page_margins(self):
        if not self._page_layout:
            self._add_page_layout()

        self._page_margins = self._page_layout.add_child(PageMargins(type_='both'))

    @property
    def score(self):
        '''
        system_layout = defaults.add_child(SystemLayout())
        system_margins = system_layout.add_child(SystemMargins())
        system_margins.add_child(LeftMargin(0))
        system_margins.add_child(RightMargin(0))
        system_layout.add_child(SystemDistance(121))
        system_layout.add_child(TopSystemDistance(300))
        '''
        return self._score

    @score.setter
    def score(self, value):
        self._score = value

    @property
    def page_height(self):
        return self._page_height

    @page_height.setter
    def page_height(self, value):
        value = self.millimeters_to_tenth(value)
        if self.page_height:
            self.page_height.value = value
        else:
            if not self._page_layout:
                self._add_page_layout()
            self._page_height = self._page_layout.add_child(PageHeight(value))

    @property
    def page_width(self):
        return self._page_width

    @page_width.setter
    def page_width(self, value):
        value = self.millimeters_to_tenth(value)
        if self.page_width:
            self.page_width.value = value
        else:
            if not self._page_layout:
                self._add_page_layout()
            self._page_width = self._page_layout.add_child(PageWidth(value))

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, value):
        if value == 'landscape':
            self.page_height, self.page_width = self.size[1], self.size[0]
        elif value == 'portrait':
            self.page_height, self.page_width = self.size[0], self.size[1]
        else:
            raise ValueError()

        self._orientation = value

    @property
    def left_margin(self):
        return self._left_margin

    @left_margin.setter
    def left_margin(self, value):
        # value = self.millimeters_to_tenth(value)
        if self.left_margin:
            self.left_margin.value = value
        else:
            if not self._page_margins:
                self._add_page_margins()
            self._left_margin = self._page_margins.add_child(LeftMargin(value))

    @property
    def right_margin(self):
        return self._right_margin

    @right_margin.setter
    def right_margin(self, value):
        # value = self.millimeters_to_tenth(value)
        if self.right_margin:
            self.right_margin.value = value
        else:
            if not self._page_margins:
                self._add_page_margins()
            self._right_margin = self._page_margins.add_child(RightMargin(value))

    @property
    def top_margin(self):
        return self._top_margin

    @top_margin.setter
    def top_margin(self, value):
        # value = self.millimeters_to_tenth(value)
        if self.top_margin:
            self.top_margin.value = value
        else:
            if not self._page_margins:
                self._add_page_margins()
            self._top_margin = self._page_margins.add_child(TopMargin(value))

    @property
    def bottom_margin(self):
        return self._bottom_margin

    @bottom_margin.setter
    def bottom_margin(self, value):
        # value = self.millimeters_to_tenth(value)
        if self.bottom_margin:
            self.bottom_margin.value = value
        else:
            if not self._page_margins:
                self._add_page_margins()
            self._bottom_margin = self._page_margins.add_child(BottomMargin(value))

    @property
    def staff_distance(self):
        return self._staff_distance

    @staff_distance.setter
    def staff_distance(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if self.staff_distance:
                self.staff_distance.value = value
            else:
                if not self._staff_layout:
                    self._add_staff_layout()
                self._staff_distance = self._staff_layout.add_child(StaffDistance(value))

    @property
    def system_distance(self):
        return self._system_distance

    @system_distance.setter
    def system_distance(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if not self._system_layout:
                self._add_system_layout()
            self.system_distance.value = value

    @property
    def system_left_margin(self):
        return self._system_left_margin

    @system_left_margin.setter
    def system_left_margin(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if not self._system_layout:
                self._add_system_layout()
            self.system_left_margin.value = value

    @property
    def system_right_margin(self):
        return self._system_right_margin

    @system_right_margin.setter
    def system_right_margin(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if not self._system_layout:
                self._add_system_layout()
            self.system_right_margin.value = value

    @property
    def top_system_distance(self):
        return self._top_system_distance

    @top_system_distance.setter
    def top_system_distance(self, value):
        if value:
            # value = self.millimeters_to_tenth(value)
            if not self._system_layout:
                self._add_system_layout()
            self.top_system_distance.value = value

    @property
    def size(self):
        (w, h) = self.sizes[self._size]
        # h = self.millimeters_to_tenth(h)
        # w = self.millimeters_to_tenth(w)
        return h, w

    @size.setter
    def size(self, value):
        self._size = value
        self.orientation = self._orientation
        # self.page_height = self.size[0]
        # self.page_width = self.size[1]

    @property
    def page_size(self):
        return self.size

    @page_size.setter
    def page_size(self, value):
        self.size = value



from musicscore.dtd.dtd import Sequence, Element, GroupReference
from musicscore.musicxml.groups.layout import Layout
from musicscore.musicxml.types.complextypes.appearance import ComplexTypeAppearance
from musicscore.musicxml.types.complextypes.complextype import ComplexType, EmptyFont
from musicscore.musicxml.types.complextypes.lyricfont import ComplexTypeLyricFont
from musicscore.musicxml.types.complextypes.lyriclanguage import ComplexTypeLyricLanguage
from musicscore.musicxml.types.complextypes.scaling import ComplexTypeScaling


class Scaling(ComplexTypeScaling):
    _TAG = 'scaling'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Appearance(ComplexTypeAppearance):
    """"""

    _TAG = 'appearance'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class MusicFont(EmptyFont):
    """"""

    _TAG = 'music-font'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class WordFont(EmptyFont):
    """"""

    _TAG = 'word-font'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class LyricFont(ComplexTypeLyricFont):
    """"""

    _TAG = 'lyric-font'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class LyricLanguage(ComplexTypeLyricLanguage):
    """"""

    _TAG = 'lyric-language'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeDefaults(ComplexType):
    """
    The defaults type specifies score-wide defaults for scaling, layout, and appearance.
    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)

    _DTD = Sequence(
        Element(Scaling, min_occurrence=0),
        GroupReference(Layout),
        Element(Appearance, 0),
        Element(MusicFont, 0),
        Element(WordFont, 0),
        Element(LyricFont, 0, None),
        Element(LyricLanguage, 0, None)
    )



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

import os
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(durations=[1, 1, 1, 1])
        for index, ch in enumerate(sf.chords):
            ch.add_words(str(index + 1))

        sf.chords[0].add_tie('start')
        sf.chords[1].add_tie('stop')
        sf.chords[0].is_adjoinable = False

        sf.chords[1].add_tie('start')
        sf.chords[2].add_tie('stop')
        sf.chords[1].is_adjoinable = False

        sf.chords[2].add_tie('start')
        sf.chords[3].add_tie('stop')
        sf.chords[2].is_adjoinable = False

        sf.chords[1].remove_tie('start')
        sf.chords[2].remove_tie('stop')

        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)

        TestScore().assert_template(xml_path)

    def test_2(self):
        sf = SimpleFormat(durations=[5], midis=[(60, 65, 69)])
        chords = sf.chords[0].split(1, 1, 1, 1, 1)
        for ch in chords:
            ch.is_adjoinable = False
        sf._chords = []
        for ch in chords:
            sf.add_chord(ch)
        sf.chords[2].remove_tie('start')
        sf.chords[3].remove_tie('stop')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_2.xml'
        self.score.write(xml_path)

        TestScore().assert_template(xml_path)

    def test_3(self):
        sf = SimpleFormat(durations=[5], midis=[(60, 65, 69)])
        chords = sf.chords[0].split(1, 1, 1, 1, 1)
        for ch in chords:
            ch.is_adjoinable = False
        sf._chords = []
        for ch in chords:
            sf.add_chord(ch)
        sf.chords[2].remove_tie('start')
        sf.chords[2].add_slur(type='start', line_type='dashed')
        sf.chords[3].remove_tie('stop')
        sf.chords[3].add_slur(type='stop')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_3.xml'
        self.score.write(xml_path)

        TestScore().assert_template(xml_path)

    def test_4(self):
        sf = SimpleFormat(durations=[5], midis=[(60, 65, 69)])
        chords = sf.chords[0].split(1, 1, 1, 1, 1)
        sf._chords = []

        for ch in chords:
            sf.add_chord(ch)
        sf.chords[2].remove_tie('start')
        sf.chords[2].add_slur(type='start', line_type='dashed')
        sf.chords[3].remove_tie('stop')
        sf.chords[3].add_slur(type='stop')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_4.xml'
        self.score.write(xml_path)

        TestScore().assert_template(xml_path)

    def test_5(self):
        sf = SimpleFormat(durations=[2])
        chords = sf.chords[0].split(1, 1)
        sf._chords = []
        for ch in chords:
            sf.add_chord(ch)
        sf.chords[0].remove_tie('start')
        sf.chords[0].add_slur(type='start', line_type='dashed')
        sf.chords[1].remove_tie('stop')
        sf.chords[1].add_slur(type='stop')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_5.xml'
        self.score.write(xml_path)

        # TestScore().assert_template(xml_path)



import os

from quicktions import Fraction

from musicscore.basic_functions import flatten
from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])
durations = [Fraction(2, 3), Fraction(279, 784), Fraction(899, 2352), Fraction(25, 42)]

durations = flatten(durations)


def get_chords():
    return score.get_measure(1).get_part(1).voices[1].chords


sf = SimpleFormat(durations=durations)
score = TreeScoreTimewise()
sf.to_stream_voice().add_to_score(score)
print('after add_to_score')
print([chord.quarter_duration for chord in get_chords()])
print(float(sum([chord.quarter_duration for chord in get_chords()])))
score.update_measures()
print('after update_measures')
print([chord.quarter_duration for chord in get_chords()])
print(float(sum([chord.quarter_duration for chord in get_chords()])))
score.fill_with_rest()
print('after fill_with_rest')
print([chord.quarter_duration for chord in get_chords()])
print(float(sum([chord.quarter_duration for chord in get_chords()])))
score.preliminary_adjoin_rests()
print('after preliminary_adjoin_rests')
print([chord.quarter_duration for chord in get_chords()])
print(float(sum([chord.quarter_duration for chord in get_chords()])))
score.add_beats()
print('after add_beats')
print([sum([chord.quarter_duration for chord in beat.chords]) for beat in score.get_beats()])
print([chord.quarter_duration for chord in get_chords()])
print(float(sum([chord.quarter_duration for chord in get_chords()])))
score.quantize()
print('after quantize')
print([chord.quarter_duration for chord in get_chords()])
print(float(sum([chord.quarter_duration for chord in get_chords()])))

xml_path = path + '_1.xml'
score.write(xml_path)



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
import os

from musicscore.musicxml.elements.note import Notations
from musicscore.musicxml.types.complextypes.harmonic import Artificial, BasePitch, TouchingPitch
from musicscore.musicxml.types.complextypes.notations import Technical
from musicscore.musicxml.types.complextypes.technical import Harmonic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(durations=[4], midis=[(60, 65)])
        chord = sf.chords[0]
        notations = chord.add_child(Notations())
        technical = notations.add_child(Technical())
        harmonic = technical.add_child(Harmonic())
        harmonic.add_child(Artificial())
        harmonic.add_child(TouchingPitch())

        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_1.xml'
        # self.score.write(xml_path)

    def test_2(self):
        sf = SimpleFormat(durations=[1.5, 2, 3, 2.33, 3.66], midis=5*[60])
        for chord in sf.chords:
            chord.add_harmonic(5)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_2.xml'
        self.score.write(xml_path)



from unittest import TestCase

from musicscore.musictree.treebeat import TreeBeat
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treepart import TreePart


class Test(TestCase):
    def setUp(self) -> None:
        self.measure = TreeMeasure()

    def test_number(self):
        self.assertTrue(isinstance(self.measure.number, str))
        with self.assertRaises(TypeError):
            self.measure.number = 3

        copied = self.measure.__copy__()
        self.assertTrue(isinstance(copied.number, str))
        with self.assertRaises(TypeError):
            copied.number = 3

    def test_duration(self):
        m = TreeMeasure(time=(3, 4))
        self.assertEqual(m.quarter_duration, 3)

        p = TreePart(id='one')
        m.add_child(p)

        tree_part_voice = p.get_voice(1)
        tree_part_voice.set_beats([TreeBeat(duration=2), TreeBeat(duration=0.5), TreeBeat(duration=0.5)])
        result = [0, 2, 2.5]
        self.assertEqual([beat.offset for beat in tree_part_voice.beats], result)



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]

#
# class Test(TestCase):
#     def setUp(self) -> None:
#         self.score = TreeScoreTimewise()
#
#     def test_1(self):
#         sf = SimpleFormat(durations=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
#         v = sf.to_stream_voice(1)
#         v.add_to_score(self.score, 1, 1)
#
#         result_path = path + '_test_1'
#         self.score.finish()
#         self.score.add_ruler()
#         self.score.write(path=result_path)
#         # TestScore().assert_template(result_path=result_path)



from unittest import TestCase
import musicscore.musicxml.elements.timewise as timewise
from musicscore.musicxml.elements.fullnote import Rest
from musicscore.musicxml.elements.note import Note, Duration
from musicscore.musicxml.groups.musicdata import Attributes
from musicscore.musicxml.types.complextypes.attributes import Divisions


class TestAddNote(TestCase):
    def setUp(self):
        self.part = timewise.Part(id='p1')
        attributes = self.part.add_child(Attributes())
        attributes.add_child(Divisions(1))

    def test_add_note(self):
        note = Note()
        note.add_child(Rest())
        note.add_child(Duration(2))
        self.part.add_child(note)

        note = Note()
        note.add_child(Rest())
        note.add_child(Duration(2))
        result = '''<part id="p1">
  <attributes>
    <divisions>1</divisions>
  </attributes>
  <note>
    <rest/>
    <duration>2</duration>
  </note>
</part>
'''
        self.assertEqual(self.part.to_string(), result)



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
import os

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(durations=[2, 2])
        sf.chords[0].add_action_dynamics('ff')
        sf.chords[1].add_dynamics('ff')

        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)



'''
	<xs:attributeGroup name="image-attributes">
		<xs:annotation>
			<xs:documentation></xs:documentation>
		</xs:annotation>
		<xs:attributeGroup ref="position"/>
		<xs:attributeGroup ref="halign"/>
		<xs:attributeGroup ref="valign-image"/>
	</xs:attributeGroup>
'''
from musicscore.musicxml.attributes.align import Halign, ValignImage
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.types.simple_type import Token


class Source(AttributeAbstract):
    def __init__(self, source, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('source', source, "AnyURI")


class Height(AttributeAbstract):
    def __init__(self, height=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('height', height, "Tenths")


class Width(AttributeAbstract):
    def __init__(self, width=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('width', width, "Tenths")


class ImageAttributes(Source, Height, Width, Position, Halign, ValignImage):
    """
    The image-attributes group is used to include graphical images in a score. The required source attribute is the URL
    for the image file. The required type attribute is the MIME type for the image file format. Typical choices include
    application/postscript, image/gif, image/jpeg, image/png, and image/tiff. The optional height and width attributes
    are used to size and scale an image. The image should be scaled independently in X and Y if both height and width
    are specified. If only one attribute is specified, the image should be scaled proportionally to fit in the
    specified dimension.
    """

    def __init__(self, _type, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = _type

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            Token(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.note import Lyric
from tests.score_templates.xml_test_score import TestScore
from musicscore.musicxml.types.complextypes.lyric import Text

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):

    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        simpleformat = SimpleFormat(midis=list(range(60, 68)))
        voice = simpleformat.to_stream_voice(2)
        self.score.add_part()
        self.score.add_measure()
        voice.add_to_score(self.score, 1)
        result_path = path + '_test_1'
        # self.score.accidental_mode = 'modern'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        simpleformat = SimpleFormat(midis=list(range(60, 68)), durations=[1.2] * 8)
        voice = simpleformat.to_stream_voice(1)
        self.score.add_part()
        self.score.add_measure()
        voice.add_to_score(self.score, 1)
        result_path = path + '_test_2'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_3(self):
        sf = SimpleFormat(midis=[(60, 61, 67)], durations=7)
        voice = sf.to_stream_voice(1)
        voice.add_to_score(self.score, 1)
        result_path = path + '_test_3'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_4(self):
        midis = list(range(60, 80))
        sf = SimpleFormat(midis=midis)
        for chord in sf.chords:
            l = chord.add_child(Lyric())
            l.add_child(Text(str(chord.midis[0].value)))
        voice = sf.to_stream_voice(1)
        voice.add_to_score(self.score, 1)
        result_path = path + '_test_4'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)
